#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:fenc=utf-8

"""
Register or fetch Pipes from the API
"""

from __future__ import annotations
import time
import json
from io import StringIO
from datetime import datetime
import meerschaum as mrsm
from meerschaum.utils.debug import dprint
from meerschaum.utils.warnings import warn, error
from meerschaum.utils.typing import SuccessTuple, Union, Any, Optional, Mapping, List, Dict, Tuple

def pipe_r_url(
        pipe: mrsm.Pipe
    ) -> str:
    """Return a relative URL path from a Pipe's keys."""
    from meerschaum.config.static import STATIC_CONFIG
    location_key = pipe.location_key
    if location_key is None:
        location_key = '[None]'
    return (
        f"{STATIC_CONFIG['api']['endpoints']['pipes']}/"
        + f"{pipe.connector_keys}/{pipe.metric_key}/{location_key}"
    )

def register_pipe(
        self,
        pipe: mrsm.Pipe,
        debug: bool = False
    ) -> SuccessTuple:
    """Submit a POST to the API to register a new Pipe object.
    Returns a tuple of (success_bool, response_dict).
    """
    from meerschaum.utils.debug import dprint
    from meerschaum.config.static import STATIC_CONFIG
    ### NOTE: if `parameters` is supplied in the Pipe constructor,
    ###       then `pipe.parameters` will exist and not be fetched from the database.
    r_url = pipe_r_url(pipe)
    response = self.post(
        r_url + '/register',
        json = pipe.parameters,
        debug = debug,
    )
    if debug:
        dprint(response.text)
    if isinstance(response.json(), list):
        response_tuple = response.__bool__(), response.json()[1]
    elif 'detail' in response.json():
        response_tuple = response.__bool__(), response.json()['detail']
    else:
        response_tuple = response.__bool__(), response.text
    return response_tuple


def edit_pipe(
        self,
        pipe: mrsm.Pipe,
        patch: bool = False,
        debug: bool = False,
    ) -> SuccessTuple:
    """Submit a PATCH to the API to edit an existing Pipe object.
    Returns a tuple of (success_bool, response_dict).
    """
    from meerschaum.utils.debug import dprint
    ### NOTE: if `parameters` is supplied in the Pipe constructor,
    ###       then `pipe.parameters` will exist and not be fetched from the database.
    r_url = pipe_r_url(pipe)
    response = self.patch(
        r_url + '/edit',
        params = {'patch': patch,},
        json = pipe.parameters,
        debug = debug,
    )
    if debug:
        dprint(response.text)
    if isinstance(response.json(), list):
        response_tuple = response.__bool__(), response.json()[1]
    elif 'detail' in response.json():
        response_tuple = response.__bool__(), response.json()['detail']
    else:
        response_tuple = response.__bool__(), response.text
    return response_tuple


def fetch_pipes_keys(
        self,
        connector_keys: Optional[List[str]] = None,
        metric_keys: Optional[List[str]] = None,
        location_keys: Optional[List[str]] = None,
        tags: Optional[List[str]] = None,
        params: Optional[Dict[str, Any]] = None,
        debug: bool = False
    ) -> Union[List[Tuple[str, str, Union[str, None]]]]:
    """
    Fetch registered Pipes' keys from the API.
    
    Parameters
    ----------
    connector_keys: Optional[List[str]], default None
        The connector keys for the query.

    metric_keys: Optional[List[str]], default None
        The metric keys for the query.

    location_keys: Optional[List[str]], default None
        The location keys for the query.

    tags: Optional[List[str]], default None
        A list of tags for the query.

    params: Optional[Dict[str, Any]], default None
        A parameters dictionary for filtering against the `pipes` table
        (e.g. `{'connector_keys': 'plugin:foo'}`).
        Not recommeded to be used.

    debug: bool, default False
        Verbosity toggle.

    Returns
    -------
    A list of tuples containing pipes' keys.
    """
    from meerschaum.config.static import STATIC_CONFIG
    if connector_keys is None:
        connector_keys = []
    if metric_keys is None:
        metric_keys = []
    if location_keys is None:
        location_keys = []
    if tags is None:
        tags = []

    r_url = STATIC_CONFIG['api']['endpoints']['pipes'] + '/keys'
    try:
        j = self.get(
            r_url,
            params = {
                'connector_keys': json.dumps(connector_keys),
                'metric_keys': json.dumps(metric_keys),
                'location_keys': json.dumps(location_keys),
                'tags': json.dumps(tags),
                'params': json.dumps(params),
            },
            debug=debug
        ).json()
    except Exception as e:
        error(str(e))

    if 'detail' in j:
        error(j['detail'], stack=False)
    return [tuple(r) for r in j]


def sync_pipe(
        self,
        pipe: mrsm.Pipe,
        df: Optional[Union['pd.DataFrame', Dict[Any, Any], str]] = None,
        chunksize: Optional[int] = -1,
        debug: bool = False,
        **kw: Any
    ) -> SuccessTuple:
    """Sync a DataFrame into a Pipe."""
    from decimal import Decimal
    from meerschaum.utils.debug import dprint
    from meerschaum.utils.misc import json_serialize_datetime, items_str
    from meerschaum.config import get_config
    from meerschaum.utils.packages import attempt_import, import_pandas
    from meerschaum.utils.dataframe import get_numeric_cols
    begin = time.time()
    more_itertools = attempt_import('more_itertools')
    if df is None:
        msg = f"DataFrame is `None`. Cannot sync {pipe}."
        return False, msg

    def get_json_str(c):
        ### allow syncing dict or JSON without needing to import pandas (for IOT devices)
        if isinstance(c, (dict, list)):
            return json.dumps(c, default=json_serialize_datetime)
        pd = import_pandas()
        return c.fillna(pd.NA).to_json(date_format='iso', date_unit='ns')

    df = json.loads(df) if isinstance(df, str) else df

    _chunksize: Optional[int] = (1 if chunksize is None else (
        get_config('system', 'connectors', 'sql', 'chunksize') if chunksize == -1
        else chunksize
    ))
    keys: List[str] = list(df.columns)
    chunks = []
    if hasattr(df, 'index'):
        df = df.reset_index(drop=True)
        is_dask = 'dask' in df.__module__
        chunks = (
            (df.iloc[i] for i in more_itertools.chunked(df.index, _chunksize))
            if not is_dask
            else [partition.compute() for partition in df.partitions]
        )
        numeric_cols = get_numeric_cols(df)
        if numeric_cols:
            for col in numeric_cols:
                df[col] = df[col].apply(lambda x: f'{x:f}' if isinstance(x, Decimal) else x)
            pipe_dtypes = pipe.dtypes
            new_numeric_cols = [
                col
                for col in numeric_cols
                if pipe_dtypes.get(col, None) != 'numeric'
            ]
            pipe.dtypes.update({
                col: 'numeric'
                for col in new_numeric_cols
            })
            edit_success, edit_msg = pipe.edit(debug=debug)
            if not edit_success:
                warn(
                    "Failed to update new numeric columns "
                    + f"{items_str(new_numeric_cols)}:\n{edit_msg}"
                )
    elif isinstance(df, dict):
        ### `_chunks` is a dict of lists of dicts.
        ### e.g. {'a' : [ {'a':[1, 2]}, {'a':[3, 4]} ] }
        _chunks = {k: [] for k in keys}
        for k in keys:
            chunk_iter = more_itertools.chunked(df[k], _chunksize)
            for l in chunk_iter:
                _chunks[k].append({k: l})

        ### `chunks` is a list of dicts (e.g. orient by rows in pandas JSON).
        for k, l in _chunks.items():
            for i, c in enumerate(l):
                try:
                    chunks[i].update(c)
                except IndexError:
                    chunks.append(c)
    elif isinstance(df, list):
        chunks = (df[i] for i in more_itertools.chunked(df, _chunksize))

    ### Send columns in case the user has defined them locally.
    if pipe.columns:
        kw['columns'] = json.dumps(pipe.columns)
    r_url = pipe_r_url(pipe) + '/data'

    rowcount = 0
    num_success_chunks = 0
    for i, c in enumerate(chunks):
        if debug:
            dprint(f"[{self}] Posting chunk {i} to {r_url}...")
        if len(c) == 0:
            if debug:
                dprint(f"[{self}] Skipping empty chunk...")
            continue
        json_str = get_json_str(c)

        try:
            response = self.post(
                r_url,
                ### handles check_existing
                params = kw,
                data = json_str,
                debug = debug
            )
        except Exception as e:
            msg = f"Failed to post a chunk to {pipe}:\n{e}"
            warn(msg)
            return False, msg
            
        if not response:
            return False, f"Failed to sync a chunk:\n{response.text}"

        try:
            j = json.loads(response.text)
        except Exception as e:
            return False, f"Failed to parse response from syncing {pipe}:\n{e}"

        if isinstance(j, dict) and 'detail' in j:
            return False, j['detail']

        try:
            j = tuple(j)
        except Exception as e:
            return False, response.text

        if debug:
            dprint("Received response: " + str(j))
        if not j[0]:
            return j

        rowcount += len(c)
        num_success_chunks += 1

    success_tuple = True, (
        f"It took {round(time.time() - begin, 2)} seconds to sync {rowcount} row"
        + ('s' if rowcount != 1 else '')
        + f" across {num_success_chunks} chunk" + ('s' if num_success_chunks != 1 else '') +
        f" to {pipe}."
    )
    return success_tuple


def delete_pipe(
        self,
        pipe: Optional[meerschaum.Pipe] = None,
        debug: bool = None,        
    ) -> SuccessTuple:
    """Delete a Pipe and drop its table."""
    if pipe is None:
        error(f"Pipe cannot be None.")
    r_url = pipe_r_url(pipe)
    response = self.delete(
        r_url + '/delete',
        debug = debug,
    )
    if debug:
        dprint(response.text)
    if isinstance(response.json(), list):
        response_tuple = response.__bool__(), response.json()[1]
    elif 'detail' in response.json():
        response_tuple = response.__bool__(), response.json()['detail']
    else:
        response_tuple = response.__bool__(), response.text
    return response_tuple


def get_pipe_data(
        self,
        pipe: meerschaum.Pipe,
        select_columns: Optional[List[str]] = None,
        omit_columns: Optional[List[str]] = None,
        begin: Union[str, datetime, int, None] = None,
        end: Union[str, datetime, int, None] = None,
        params: Optional[Dict[str, Any]] = None,
        as_chunks: bool = False,
        debug: bool = False,
        **kw: Any
    ) -> Union[pandas.DataFrame, None]:
    """Fetch data from the API."""
    r_url = pipe_r_url(pipe)
    chunks_list = []
    while True:
        try:
            response = self.get(
                r_url + "/data",
                params = {
                    'select_columns': json.dumps(select_columns),
                    'omit_columns': json.dumps(omit_columns),
                    'begin': begin,
                    'end': end,
                    'params': json.dumps(params)
                },
                debug = debug
            )
            if not response.ok:
                return None
            j = response.json()
        except Exception as e:
            warn(f"Failed to get data for {pipe}:\n{e}")
            return None
        if isinstance(j, dict) and 'detail' in j:
            return False, j['detail']
        break

    from meerschaum.utils.packages import import_pandas
    from meerschaum.utils.dataframe import parse_df_datetimes, add_missing_cols_to_df
    pd = import_pandas()
    try:
        df = pd.read_json(StringIO(response.text))
    except Exception as e:
        warn(f"Failed to parse response for {pipe}:\n{e}")
        return None

    if len(df.columns) == 0:
        return add_missing_cols_to_df(df, pipe.dtypes)

    df = parse_df_datetimes(
        df,
        ignore_cols = [
            col
            for col, dtype in pipe.dtypes.items()
            if 'datetime' not in str(dtype)
        ],
        debug = debug,
    )
    return df


def get_pipe_id(
        self,
        pipe: meerschuam.Pipe,
        debug: bool = False,
    ) -> int:
    """Get a Pipe's ID from the API."""
    from meerschaum.utils.misc import is_int
    r_url = pipe_r_url(pipe)
    response = self.get(
        r_url + '/id',
        debug = debug
    )
    if debug:
        dprint(f"Got pipe ID: {response.text}")
    try:
        if is_int(response.text):
            return int(response.text)
    except Exception as e:
        warn(f"Failed to get the ID for {pipe}:\n{e}")
    return None


def get_pipe_attributes(
        self,
        pipe: meerschaum.Pipe,
        debug: bool = False,
    ) -> Dict[str, Any]:
    """Get a Pipe's attributes from the API

    Parameters
    ----------
    pipe: meerschaum.Pipe
        The pipe whose attributes we are fetching.
        
    Returns
    -------
    A dictionary of a pipe's attributes.
    If the pipe does not exist, return an empty dictionary.
    """
    r_url = pipe_r_url(pipe)
    response = self.get(r_url + '/attributes', debug=debug)
    try:
        return json.loads(response.text)
    except Exception as e:
        warn(f"Failed to get the attributes for {pipe}:\n{e}")
    return {}


def get_sync_time(
        self,
        pipe: 'meerschaum.Pipe',
        params: Optional[Dict[str, Any]] = None,
        newest: bool = True,
        debug: bool = False,
    ) -> Union[datetime, int, None]:
    """Get a Pipe's most recent datetime value from the API.

    Parameters
    ----------
    pipe: meerschaum.Pipe
        The pipe to select from.

    params: Optional[Dict[str, Any]], default None
        Optional params dictionary to build the WHERE clause.

    newest: bool, default True
        If `True`, get the most recent datetime (honoring `params`).
        If `False`, get the oldest datetime (ASC instead of DESC).

    Returns
    -------
    The most recent (or oldest if `newest` is `False`) datetime of a pipe,
    rounded down to the closest minute.
    """
    from meerschaum.utils.misc import is_int
    from meerschaum.utils.warnings import warn
    r_url = pipe_r_url(pipe)
    response = self.get(
        r_url + '/sync_time',
        json = params,
        params = {'newest': newest, 'debug': debug},
        debug = debug,
    )
    if not response:
        warn(f"Failed to get the sync time for {pipe}:\n" + response.text)
        return None

    j = response.json()
    if j is None:
        dt = None
    else:
        try:
            dt = (
                datetime.fromisoformat(j)
                if not is_int(j)
                else int(j)
            )
        except Exception as e:
            warn(f"Failed to parse the sync time '{j}' for {pipe}:\n{e}")
            dt = None
    return dt


def pipe_exists(
        self,
        pipe: 'meerschaum.Pipe',
        debug: bool = False
    ) -> bool:
    """Check the API to see if a Pipe exists.

    Parameters
    ----------
    pipe: 'meerschaum.Pipe'
        The pipe which were are querying.
        
    Returns
    -------
    A bool indicating whether a pipe's underlying table exists.
    """
    from meerschaum.utils.debug import dprint
    from meerschaum.utils.warnings import warn
    r_url = pipe_r_url(pipe)
    response = self.get(r_url + '/exists', debug=debug)
    if not response:
        warn(f"Failed to check if {pipe} exists:\n{response.text}")
        return False
    if debug:
        dprint("Received response: " + str(response.text))
    j = response.json()
    if isinstance(j, dict) and 'detail' in j:
        warn(j['detail'])
    return j


def create_metadata(
        self,
        debug: bool = False
    ) -> bool:
    """Create metadata tables.

    Returns
    -------
    A bool indicating success.
    """
    from meerschaum.utils.debug import dprint
    from meerschaum.config.static import STATIC_CONFIG
    r_url = STATIC_CONFIG['api']['endpoints']['metadata']
    response = self.post(r_url, debug=debug)
    if debug:
        dprint("Create metadata response: {response.text}")
    try:
        metadata_response = json.loads(response.text)
    except Exception as e:
        warn(f"Failed to create metadata on {self}:\n{e}")
        metadata_response = False
    return False


def get_pipe_rowcount(
        self,
        pipe: 'meerschaum.Pipe',
        begin: Optional[datetime] = None,
        end: Optional[datetime] = None,
        params: Optional[Dict[str, Any]] = None,
        remote: bool = False,
        debug: bool = False,
    ) -> int:
    """Get a pipe's row count from the API.

    Parameters
    ----------
    pipe: 'meerschaum.Pipe':
        The pipe whose row count we are counting.
        
    begin: Optional[datetime], default None
        If provided, bound the count by this datetime.

    end: Optional[datetime]
        If provided, bound the count by this datetime.

    params: Optional[Dict[str, Any]], default None
        If provided, bound the count by these parameters.

    remote: bool, default False

    Returns
    -------
    The number of rows in the pipe's table, bound the given parameters.
    If the table does not exist, return 0.
    """
    r_url = pipe_r_url(pipe)
    response = self.get(
        r_url + "/rowcount",
        json = params,
        params = {
            'begin': begin,
            'end': end,
            'remote': remote,
        },
        debug = debug
    )
    if not response:
        warn(f"Failed to get the rowcount for {pipe}:\n{response.text}")
        return 0
    try:
        return int(json.loads(response.text))
    except Exception as e:
        warn(f"Failed to get the rowcount for {pipe}:\n{e}")
    return 0


def drop_pipe(
        self,
        pipe: meerschaum.Pipe,
        debug: bool = False
    ) -> SuccessTuple:
    """
    Drop a pipe's table but maintain its registration.

    Parameters
    ----------
    pipe: meerschaum.Pipe:
        The pipe to be dropped.
        
    Returns
    -------
    A success tuple (bool, str).
    """
    from meerschaum.utils.warnings import error
    from meerschaum.utils.debug import dprint
    if pipe is None:
        error(f"Pipe cannot be None.")
    r_url = pipe_r_url(pipe)
    response = self.delete(
        r_url + '/drop',
        debug = debug,
    )
    if debug:
        dprint(response.text)

    try:
        data = response.json()
    except Exception as e:
        return False, f"Failed to drop {pipe}."

    if isinstance(data, list):
        response_tuple = response.__bool__(), data[1]
    elif 'detail' in response.json():
        response_tuple = response.__bool__(), data['detail']
    else:
        response_tuple = response.__bool__(), response.text

    return response_tuple


def clear_pipe(
        self,
        pipe: meerschaum.Pipe,
        debug: bool = False,
        **kw
    ) -> SuccessTuple:
    """
    Delete rows in a pipe's table.

    Parameters
    ----------
    pipe: meerschaum.Pipe
        The pipe with rows to be deleted.
        
    Returns
    -------
    A success tuple.
    """
    kw.pop('metric_keys', None)
    kw.pop('connector_keys', None)
    kw.pop('location_keys', None)
    kw.pop('action', None)
    kw.pop('force', None)
    return self.do_action(
        ['clear', 'pipes'],
        connector_keys = pipe.connector_keys,
        metric_keys = pipe.metric_key,
        location_keys = pipe.location_key,
        force = True,
        debug = debug,
        **kw
    )


def get_pipe_columns_types(
        self,
        pipe: meerschaum.Pipe,
        debug: bool = False,
    ) -> Union[Dict[str, str], None]:
    """
    Fetch the columns and types of the pipe's table.

    Parameters
    ----------
    pipe: meerschaum.Pipe
        The pipe whose columns to be queried.
        
    Returns
    -------
    A dictionary mapping column names to their database types.

    Examples
    --------
    >>> {
    ...   'dt': 'TIMESTAMP WITHOUT TIMEZONE',
    ...   'id': 'BIGINT',
    ...   'val': 'DOUBLE PRECISION',
    ... }
    >>>
    """
    r_url = pipe_r_url(pipe) + '/columns/types'
    response = self.get(
        r_url,
        debug = debug
    )
    j = response.json()
    if isinstance(j, dict) and 'detail' in j and len(j.keys()) == 1:
        from meerschaum.utils.warnings import warn
        warn(j['detail'])
        return None
    if not isinstance(j, dict):
        warn(response.text)
        return None
    return j



#! /usr/bin/env python3
# -*- coding: utf-8 -*-
# vim:fenc=utf-8

"""
Pause running jobs that were started with `-d` or `start job`.
"""

from __future__ import annotations
from meerschaum.utils.typing import Optional, List, Dict, SuccessTuple, Any

def pause(action: Optional[List[str]] = None, **kw) -> SuccessTuple:
    """
    Pause running jobs.
    """
    from meerschaum.actions import choose_subaction
    options = {
        'jobs': _pause_jobs,
    }
    return choose_subaction(action, options, **kw)


def _complete_pause(
        action: Optional[List[str]] = None,
        **kw: Any
    ) -> List[str]:
    """
    Override the default Meerschaum `complete_` function.
    """
    from meerschaum.actions.start import _complete_start_jobs
    if action is None:
        action = []

    options = {
        'job' : _complete_start_jobs,
        'jobs' : _complete_start_jobs,
    }

    if (
        len(action) > 0 and action[0] in options
            and kw.get('line', '').split(' ')[-1] != action[0]
    ):
        sub = action[0]
        del action[0]
        return options[sub](action=action, **kw)

    from meerschaum._internal.shell import default_action_completer
    return default_action_completer(action=(['start'] + action), **kw)


def _pause_jobs(
        action: Optional[List[str]] = None,
        noask: bool = False,
        force: bool = False,
        yes: bool = False,
        nopretty: bool = False,
        **kw
    ) -> SuccessTuple:
    """
    Pause running jobs that were started with `-d` or `start job`.
    
    To see running jobs, run `show jobs`.
    """
    from meerschaum.utils.formatting._jobs import pprint_jobs
    from meerschaum.utils.daemon import (
        get_filtered_daemons, get_running_daemons, get_stopped_daemons, get_paused_daemons,
    )
    from meerschaum.utils.warnings import warn
    from meerschaum.utils.prompt import yes_no
    from meerschaum.utils.misc import items_str
    daemons = get_filtered_daemons(action, warn=(not nopretty))
    _running_daemons = get_running_daemons(daemons)
    _paused_daemons = get_paused_daemons(daemons)
    _stopped_daemons = get_stopped_daemons(daemons)
    if action and _stopped_daemons and not nopretty:
        warn(
            f"Skipping stopped job" + ("s" if len(_stopped_daemons) > 1 else '') + " '" +
                ("', '".join(d.daemon_id for d in _stopped_daemons)) + "'.",
            stack = False
        )

    daemons_to_pause = _running_daemons + _paused_daemons
    if not daemons_to_pause:
        return False, "No running jobs to pause. You can start jobs with `-d` or `start jobs`."

    if not action:
        if not force:
            pprint_jobs(_running_daemons)
            if not yes_no(
                "Pause all running jobs?",
                noask=noask, yes=yes, default='n'
            ):
                return False, "No jobs were paused."

    successes, fails = [], []
    for d in daemons_to_pause:
        pause_success, pause_msg = d.pause()
        (successes if pause_success else fails).append(d.daemon_id)

    msg = (
        (
            (
                "Successfully paused job" + ("s" if len(successes) != 1 else '')
                + f" {items_str(successes)}."
                + (
                    '\n'
                    if fails
                    else ''
                )
            )
            if successes
            else ''
        )
        + (
            "Failed to pause job" + ("s" if len(fails) != 1 else '')
            + f" {items_str(fails)}."
            if fails
            else ''
        )
    )
    return len(fails) == 0, msg


### NOTE: This must be the final statement of the module.
###       Any subactions added below these lines will not
###       be added to the `help` docstring.
from meerschaum.actions import choices_docstring as _choices_docstring
pause.__doc__ += _choices_docstring('pause')



#! /usr/bin/env python3
# -*- coding: utf-8 -*-
# vim:fenc=utf-8

"""
Callbacks for the plugins page.
"""

from __future__ import annotations
import json
from meerschaum.utils.typing import Optional, Dict, Any
from meerschaum.api.dash import dash_app, debug, active_sessions
from meerschaum.api import get_api_connector, endpoints, CHECK_UPDATE
from meerschaum.core import Plugin
from meerschaum.utils.packages import attempt_import, import_dcc, import_html
dash = attempt_import('dash', lazy=False, check_update=CHECK_UPDATE)
from dash.exceptions import PreventUpdate
from dash.dependencies import Input, Output, State, ALL, MATCH
html, dcc = import_html(check_update=CHECK_UPDATE), import_dcc(check_update=CHECK_UPDATE)
import dash_bootstrap_components as dbc
from meerschaum.api.dash.components import alert_from_success_tuple, build_cards_grid
from dash.exceptions import PreventUpdate
from meerschaum.api.dash.plugins import get_plugins_cards, is_plugin_owner

@dash_app.callback(
    Output('plugins-cards-div', 'children'),
    Input('search-plugins-input', 'value'),
    State('session-store', 'data'),
)
def search_plugins(text: Optional[str] = None, session_data: Optional[Dict[str, Any]] = None):
    cards, alerts = get_plugins_cards(search_term=text, session_data=session_data)
    return build_cards_grid(cards, num_columns=3)


@dash_app.callback(
    Output({'type': 'edit-alert-div', 'index': MATCH}, 'children'),
    Input({'type': 'edit-button', 'index': MATCH}, 'n_clicks'),
    State({'type': 'description-textarea', 'index': MATCH}, 'value'),
    State('session-store', 'data'),
)
def edit_plugin_description(
    n_clicks: Optional[int] = None,
    description: Optional[str] = None,
    session_data: Optional[Dict[str, Any]] = None,
):
    """
    Edit a plugin's description and set the alert.
    """
    if n_clicks is None:
        raise PreventUpdate
    ctx = dash.callback_context
    prop_id = ctx.triggered[0]['prop_id']
    j = '.'.join(prop_id.split('.')[:-1])
    plugin_name = json.loads(j)['index']
    if not is_plugin_owner(plugin_name, session_data):
        success, msg = (
            False,
            f"Failed to update description for plugin '{plugin_name}'"
            + " due to insufficient permissions."
        )
    else:
        plugin = Plugin(plugin_name)
        plugin.user_id = get_api_connector().get_plugin_user_id(plugin, debug=debug)
        plugin.attributes = get_api_connector().get_plugin_attributes(plugin, debug=debug)
        plugin.attributes.update({'description': description})
        success, _msg = get_api_connector().register_plugin(plugin, debug=debug, force=True)
        msg = _msg if not success else "Successfully updated description."
    return [alert_from_success_tuple((success, msg))]




#! /usr/bin/env python3
# -*- coding: utf-8 -*-
# vim:fenc=utf-8

"""
Verify the states of pipes, pacakages, and more.
"""

from __future__ import annotations
from meerschaum.utils.typing import Union, Any, Sequence, SuccessTuple, Optional, Tuple, List

def verify(
        action: Optional[List[str]] = None,
        **kwargs: Any
    ) -> SuccessTuple:
    """
    Verify the states of pipes, packages, and more.
    """
    from meerschaum.actions import choose_subaction
    options = {
        'pipes': _verify_pipes,
        'packages': _verify_packages,
        'venvs': _verify_venvs,
        'plugins': _verify_plugins,
    }
    return choose_subaction(action, options, **kwargs)


def _verify_pipes(**kwargs) -> SuccessTuple:
    """
    Verify the contents of pipes, syncing across their entire datetime intervals.
    """
    from meerschaum.actions.sync import _sync_pipes
    kwargs['verify'] = True
    return _sync_pipes(**kwargs)


def _verify_packages(
        debug: bool = False,
        venv: Optional[str] = 'mrsm',
        **kw
    ) -> SuccessTuple:
    """
    Verify the versions of packages.
    """
    from meerschaum.utils.packages import (
        attempt_import, need_update, all_packages, is_installed, venv_contains_package,
        _monkey_patch_get_distribution, manually_import_module,
    )
    from meerschaum.utils.formatting import pprint
    from meerschaum.utils.debug import dprint

    venv_packages, base_packages, miss_packages = [], [], []

    ### Verify the system dependencies.
    for import_name, install_name in all_packages.items():
        _where_list = (
            venv_packages if venv_contains_package(
                import_name, split=False, venv=venv, debug=debug
            ) else (
                base_packages if is_installed(import_name, venv=None)
                else miss_packages
            )
        )
        _where_list.append(import_name)

    if 'flask_compress' in venv_packages or 'dash' in venv_packages:
        flask_compress = attempt_import('flask_compress', lazy=False, debug=debug)
        _monkey_patch_get_distribution('flask-compress', flask_compress.__version__)
        if 'flask_compress' in venv_packages:
            venv_packages.remove('flask_compress')

    for import_name in base_packages:
        manually_import_module(import_name, debug=debug, venv=None)
    for import_name in venv_packages:
        manually_import_module(import_name, debug=debug, venv=venv)

    ### Verify the plugins dependencies.
    return True, f"Verified {len(base_packages) + len(venv_packages)} packages."


def _verify_venvs(
        action: Optional[List[str]],
        debug: bool = False,
        **kw
    ) -> SuccessTuple:
    """
    Verify your virtual environments.
    """
    from meerschaum.utils.venv import get_venvs, verify_venv
    for venv in get_venvs():
        verify_venv(venv)
    return True, "Success"


def _verify_plugins(
        action: Optional[List[str]] = None,
        **kwargs: Any
    ) -> SuccessTuple:
    """
    Verify that all of the available plugins are able to be imported as expected.
    """
    from meerschaum.utils.formatting import print_options, UNICODE, print_tuple
    from meerschaum.plugins import import_plugins, get_plugins_names, Plugin
    from meerschaum.config import get_config
    from meerschaum.utils.misc import items_str

    plugins_names_to_verify = action or get_plugins_names()
    if not plugins_names_to_verify:
        if not action:
            return True, "There are no installed plugins; nothing to do."
        return False, f"Unable to verify plugins {items_str(action)}."

    header = f"Verifying {len(plugins_names_to_verify)} plugins"
    print_options(
        plugins_names_to_verify,
        header = header,
    )

    failed_to_import = []
    for plugin_name in plugins_names_to_verify:
        plugin = Plugin(plugin_name)
        plugin_success = plugin.module is not None
        if not plugin_success:
            failed_to_import.append(plugin)
        plugin_msg = (
            (f"Imported plugin '{plugin}'.")
            if plugin_success
            else (f"Failed to import plugin '{plugin}'.")
        )
        print_tuple((plugin_success, plugin_msg), calm=True)

    success = len(failed_to_import) == 0
    message = (
        f"Successfully imported {len(plugins_names_to_verify)} plugins."
        if success
        else (
            f"Failed to import plugin"
            + ('s' if len(failed_to_import) != 1 else '')
            + f" {items_str(failed_to_import)}."
        )
    )
    return success, message


### NOTE: This must be the final statement of the module.
###       Any subactions added below these lines will not
###       be added to the `help` docstring.
from meerschaum.actions import choices_docstring as _choices_docstring
verify.__doc__ += _choices_docstring('verify')



#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:fenc=utf-8

"""
The main dashboard layout.
"""

import uuid
from meerschaum.config import __doc__ as doc, get_config
from meerschaum.utils.misc import get_connector_labels
from meerschaum.utils.packages import attempt_import, import_html, import_dcc, import_pandas
from meerschaum.api import endpoints, CHECK_UPDATE
dex = attempt_import('dash_extensions', lazy=False, check_update=CHECK_UPDATE)
dbc = attempt_import('dash_bootstrap_components', lazy=False, check_update=CHECK_UPDATE)
html, dcc = import_html(check_update=CHECK_UPDATE), import_dcc(check_update=CHECK_UPDATE)
pd = import_pandas(check_update=CHECK_UPDATE)
px = attempt_import('plotly.express', warn=False, check_update=CHECK_UPDATE)
daq = attempt_import('dash_daq', warn=False, check_update=CHECK_UPDATE)

from meerschaum.api.dash.components import (
    go_button,
    search_parameters_editor,
    test_button,
    get_items_menu,
    bottom_buttons_content,
    console_div,
    download_dataframe,
    navbar,
    download_logs,
    refresh_jobs_interval,
)
from meerschaum.api.dash.keys import (
    keys_lists_content,
    text_tab_content,
    dropdown_tab_content,
)

layout = html.Div(
    id = 'main-div',
    children = [
        keys_lists_content,
        download_dataframe,
        download_logs,
        refresh_jobs_interval,
        navbar,
        html.Div(
            dbc.Row(
                id = 'content-row',
                children = [
                    dbc.Col(
                        children = [
                            dbc.Tabs(
                                id = 'pipes-filter-tabs',
                                children = [
                                    dbc.Tab(
                                        dropdown_tab_content,
                                        label = 'Filter',
                                        id = 'pipes-filter-dropdown-tab',
                                        tab_id = 'dropdown',
                                    ),
                                    dbc.Tab(
                                        text_tab_content,
                                        label = 'Text',
                                        id = 'pipes-filter-input-tab',
                                        tab_id = 'input',
                                        tab_style = {"display": "none"},
                                    ),
                                ]
                            ),
                            html.Br(),
                            bottom_buttons_content,
                            test_button,
                            html.Div(id='ws-div'),
                        ],
                        id = 'content-col-left',
                        md = 12,
                        lg = 6,
                    ),
                    dbc.Col(
                        children = [
                            dbc.Col([
                                    html.Div(id='success-alert-div'),
                                    html.Div(id='instance-alert-div')
                                ],
                                width={'size': 8, 'offset': 2}
                            ),
                            html.Div(id='webterm-div'),
                            html.Div(
                                id = 'content-div-right',
                                children = [console_div],
                            ),
                            html.Div(id='terminal'),
                        ],
                        md = 12,
                        lg = 6,
                        id = 'content-col-right',
                    ),
                ],
                style = {'max-width': '100%', 'padding': '15px'},
            ), ### end of Row
            className = 'container-fluid',
        ), ### end of Div
        html.P('', id='line-buffer', style = {'display': 'none'}),
    ],
)



#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:fenc=utf-8
"""
Functions for dropping elements
"""

from __future__ import annotations
from meerschaum.utils.typing import SuccessTuple, Union, Any, Optional, Sequence, List

def drop(
        action: Optional[List[str]] = None,
        **kw: Any
    ) -> SuccessTuple:
    """
    Drop pipe data (maintaining registration) or tables.
    """
    from meerschaum.actions import choose_subaction
    options = {
        'pipes'  : _drop_pipes,
    }
    return choose_subaction(action, options, **kw)


def _drop_pipes(
        action: Optional[List[str]] = None,
        yes: bool = False,
        force: bool = False,
        noask: bool = False,
        debug: bool = False,
        **kw: Any
    ) -> SuccessTuple:
    """
    Drop pipes' tables but keep pipe metadata registration.
    """
    from meerschaum.utils.prompt import prompt, yes_no
    from meerschaum import get_pipes
    from meerschaum.utils.warnings import warn
    from meerschaum.utils.debug import dprint
    from meerschaum.utils.formatting import pprint
    pipes = get_pipes(as_list=True, debug=debug, **kw)
    if len(pipes) == 0:
        return False, "No pipes to drop."
    question = (
        "Are you sure you want to drop these tables?\n    "
        + "Data will be lost and will need to be resynced.\n\n"
    )
    for p in pipes:
        question += f"    - {p.target}" + "\n"
    question += '\n'
    if force:
        answer = True
    else:
        answer = yes_no(question, default='n', noask=noask, yes=yes)
    if not answer:
        return False, "No pipes were dropped."

    success_dict = {}
    successes, fails = 0, 0
    msg = ""

    for p in pipes:
        success_tup = p.drop(debug=debug)
        success_dict[p] = success_tup[1]
        if success_tup[0]:
            successes += 1
        else:
            fails += 1
            warn(success_tup[1], stack=False)
    
    if debug:
        dprint("Results for dropping pipes.")
        pprint(success_dict)

    msg = (
        f"Finished dropping {len(pipes)} pipes\n" + 
        f"    ({successes} succeeded, {fails} failed)."
    )
    return successes > 0, msg


### NOTE: This must be the final statement of the module.
###       Any subactions added below these lines will not
###       be added to the `help` docstring.
from meerschaum.actions import choices_docstring as _choices_docstring
drop.__doc__ += _choices_docstring('drop')



#! /usr/bin/env python3
# -*- coding: utf-8 -*-
# vim:fenc=utf-8

"""
Stop running jobs that were started with `-d` or `start job`.
"""

from __future__ import annotations
from meerschaum.utils.typing import Optional, List, Dict, SuccessTuple, Any

def stop(action: Optional[List[str]] = None, **kw) -> SuccessTuple:
    """
    Stop running jobs.
    """
    from meerschaum.actions import choose_subaction
    options = {
        'jobs': _stop_jobs,
    }
    return choose_subaction(action, options, **kw)


def _complete_stop(
        action: Optional[List[str]] = None,
        **kw: Any
    ) -> List[str]:
    """
    Override the default Meerschaum `complete_` function.
    """
    from meerschaum.actions.start import _complete_start_jobs
    if action is None:
        action = []

    options = {
        'job' : _complete_start_jobs,
        'jobs' : _complete_start_jobs,
    }

    if (
        len(action) > 0 and action[0] in options
            and kw.get('line', '').split(' ')[-1] != action[0]
    ):
        sub = action[0]
        del action[0]
        return options[sub](action=action, **kw)

    from meerschaum._internal.shell import default_action_completer
    return default_action_completer(action=(['start'] + action), **kw)


def _stop_jobs(
        action: Optional[List[str]] = None,
        timeout_seconds: Optional[int] = None,
        noask: bool = False,
        force: bool = False,
        yes: bool = False,
        nopretty: bool = False,
        **kw
    ) -> SuccessTuple:
    """
    Stop running jobs that were started with `-d` or `start job`.
    
    To see running processes, run `show jobs`.
    """
    from meerschaum.utils.formatting._jobs import pprint_jobs
    from meerschaum.utils.daemon import (
        get_filtered_daemons, get_running_daemons, get_stopped_daemons, get_paused_daemons,
    )
    from meerschaum.utils.warnings import warn
    from meerschaum.utils.prompt import yes_no
    daemons = get_filtered_daemons(action, warn=(not nopretty))
    _running_daemons = get_running_daemons(daemons)
    _paused_daemons = get_paused_daemons(daemons)
    _stopped_daemons = get_stopped_daemons(daemons)
    if action and _stopped_daemons and not nopretty:
        warn(
            f"Skipping stopped job" + ("s" if len(_stopped_daemons) > 1 else '') + " '" +
                ("', '".join(d.daemon_id for d in _stopped_daemons)) + "'.",
            stack = False
        )

    daemons_to_stop = _running_daemons + _paused_daemons
    if not daemons_to_stop:
        return False, "No running or paused jobs to stop."

    if not action:
        if not force:
            pprint_jobs(daemons_to_stop)
            if not yes_no(
                "Stop the above jobs?",
                noask=noask, yes=yes, default='n'
            ):
                return False, "No jobs were stopped."

    _quit_daemons, _kill_daemons = [], []
    for d in daemons_to_stop:
        quit_success, quit_msg = d.quit(timeout=timeout_seconds)
        if quit_success:
            _quit_daemons.append(d)
            continue
        else:
            warn(
                f"Failed to gracefully quit job '{d.daemon_id}', attempting to terminate:\n   "
                + f"{quit_msg}",
                stack = False,
            )

        kill_success, kill_msg = d.kill(timeout=timeout_seconds)
        if kill_success:
            _kill_daemons.append(d)
            continue
        if not nopretty:
            warn(f"Failed to kill job '{d.daemon_id}' (PID {d.pid}):\n{kill_msg}", stack=False)

    msg = (
        (("Stopped job" + ("s" if len(_quit_daemons) != 1 else '') +
            " '" + "', '".join([d.daemon_id for d in _quit_daemons]) + "'.")
            if _quit_daemons else '')
        + (("\n" if _quit_daemons else "")
           + ("Killed job" + ("s" if len(_kill_daemons) != 1 else '') +
            " '" + "', '".join([d.daemon_id for d in _kill_daemons]) + "'.")
            if _kill_daemons else '')
    )
    return (len(_quit_daemons + _kill_daemons) > 0), msg


### NOTE: This must be the final statement of the module.
###       Any subactions added below these lines will not
###       be added to the `help` docstring.
from meerschaum.actions import choices_docstring as _choices_docstring
stop.__doc__ += _choices_docstring('stop')



#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:fenc=utf-8
"""
Import the config yaml file
"""

from __future__ import annotations
from meerschaum.utils.typing import Optional, Dict, Any, List, Tuple, Union
from meerschaum.config import get_config

def read_config(
        directory: Optional[str] = None,
        keys: Optional[List[str]] = None,
        write_missing : bool = True,
        substitute : bool = True,
        with_filenames : bool = False,
    ) -> Union[Dict[str, Any], Tuple[Dict[str, Any], List[str]]]:
    """
    Read the configuration directory.

    Parameters
    ----------
    directory: Optional[str], default None
        The directory with configuration files (.json and .yaml).

    keys: Optional[List[str]], default None
        Which configuration files to read.

    write_missing: bool, default True
        If a keyfile does not exist but is defined in the default configuration,
        write the file to disk.

    substitute: bool, default True
        Replace `MRSM{}` syntax with configuration values.

    with_filename: bool, default False
        If `True`, return a tuple of the configuration dictionary with a list of read filenames.
        
    Examples
    --------
    >>> read_config(keys=['meerschaum'], with_filename=True)
    >>> ({...}, ['meerschaum.yaml'])
    """
    import sys, shutil, os, json, itertools
    from meerschaum.utils.packages import attempt_import
    from meerschaum.utils.yaml import yaml, _yaml
    from meerschaum.config._paths import CONFIG_DIR_PATH
    from meerschaum.config.static import STATIC_CONFIG
    from meerschaum.config._patch import apply_patch_to_config
    if directory is None:
        directory = CONFIG_DIR_PATH

    if _yaml is None:
        print('Could not import YAML! Reverting to default configuration.')
        from meerschaum.config._default import default_config
        return default_config

    ### Each key corresponds to a YAML or JSON file.
    symlinks_key = STATIC_CONFIG['config']['symlinks_key']
    config = {}
    config_to_write = {}

    default_filetype = STATIC_CONFIG['config']['default_filetype']
    filetype_loaders = {
        'yml' : yaml.load,
        'yaml' : yaml.load,
        'json' : json.load,
    }

    ### Construct filekeys (files to parse).
    filekeys = []
    filenames = os.listdir(directory)
    missing_keys, found_keys = set(), set()
    if keys is None:
        _filekeys = filenames
    else:
        _filekeys = []
        for k in keys:
            for ft in filetype_loaders:
                if str(k) + '.' + str(ft) in filenames:
                    _filekeys.append(str(k) + '.' + str(ft))
                    found_keys.add(k)
                    if k in missing_keys:
                        missing_keys.remove(k)
                elif k not in found_keys:
                    missing_keys.add(k)

    ### Check for missing files with default keys.
    if len(missing_keys) > 0:
        from meerschaum.config._default import default_config
        for mk in missing_keys:
            if mk not in default_config:
                continue
            _default_dict = (
                search_and_substitute_config(default_config) if substitute
                else default_config
            )
            ### If default config contains symlinks, add them to the config to write.
            try:
                _default_symlinks = _default_dict[symlinks_key][mk]
            except Exception as e:
                _default_symlinks = {}
            config[mk] = _default_dict[mk]
            if _default_symlinks:
                if symlinks_key not in config:
                    config[symlinks_key] = {}
                if mk not in config[symlinks_key]:
                    config[symlinks_key][mk] = {}
                config[symlinks_key][mk] = apply_patch_to_config(
                    config[symlinks_key][mk], 
                    _default_symlinks
                )
                if symlinks_key not in config_to_write:
                    config_to_write[symlinks_key] = {}
                config_to_write[symlinks_key][mk] = config[symlinks_key][mk]

            ### Write the default key.
            config_to_write[mk] = config[mk]

    ### Write missing keys if necessary.
    if len(config_to_write) > 0 and write_missing:
        from meerschaum.config._edit import write_config
        write_config(config_to_write, directory)

    ### Check for duplicate files.
    ### Found help on StackOverflow:
    ### https://stackoverflow.com/questions/26618688/python-iterate-over-a-list-
    ### of-files-finding-same-filenames-but-different-exten
    keygroups = {
        key: list(value)
        for key, value in itertools.groupby(
            sorted(_filekeys, key = lambda e: os.path.splitext(e)[0]),
            key = lambda e: os.path.splitext(e)[0]
        )
    }
    for k, v in keygroups.items():
        fn = v[0]
        if len(v) > 1:
            if k + '.' + default_filetype in v:
                fn = k + '.' + default_filetype
            print(
                f"Found multiple config files named '{k}'. " +
                f"Will attempt to parse '{fn}' for key '{k}'."
            )
        filekeys.append(fn)

    _seen_keys = []
    for filename in filekeys:
        filepath = os.path.join(directory, filename)
        _parts = filename.split('.')
        _type = _parts[-1]
        key = '.'.join(_parts[:-1])
        ### Check if we've seen this key before (e.g. test.yaml, test.yml, test.json).
        if key in _seen_keys:
            print(
                f"Multiple files with the name '{key}' found in '{str(directory)}'. " +
                f"Reading from '{filename}'."
            )
        if len(_parts) < 2 or _type not in filetype_loaders:
            print(f"Unknown file '{filename}' in '{str(directory)}'. Skipping...")

        while True:
            try:
                with open(filepath, 'r', encoding='utf-8') as f:
                    try:
                        _config_key = filetype_loaders[_type](f)
                    except Exception as e:
                        print(f"Error processing file: {filepath}")
                        import traceback
                        traceback.print_exc()
                        _config_key = {}
                _single_key_config = (
                    search_and_substitute_config({key: _config_key}) if substitute
                    else {key: _config_key}
                )
                config[key] = _single_key_config[key]
                if (
                    symlinks_key in _single_key_config
                    and key in _single_key_config[symlinks_key]
                ):
                    if symlinks_key not in config:
                        config[symlinks_key] = {}
                    config[symlinks_key][key] = _single_key_config[symlinks_key][key]
                break
            except Exception as e:
                print(f"Unable to parse {filename}!")
                import traceback
                traceback.print_exc()
                input(f"Press [Enter] to open '{filename}' and fix formatting errors.")
                from meerschaum.utils.misc import edit_file
                edit_file(filepath)

    if with_filenames:
        return config, filekeys
    return config


def search_and_substitute_config(
        config: Dict[str, Any],
        leading_key: str = "MRSM",
        delimiter: str = ":",
        begin_key: str = "{",
        end_key: str = "}",
        literal_key: str = '!',
        keep_symlinks: bool = True,
    ) -> Dict[str, Any]:
    """Search the config for Meerschaum substitution syntax and substite with value of keys.

    Parameters
    ----------
    config: Dict[str, Any]
        The Meerschaum configuration dictionary to search through.

    leading_key: str, default 'MRSM'
        The string with which to start the search.

    begin_key: str, default '{'
        The string to start the keys list.

    end_key: str, default '}'
        The string to end the keys list.

    literal_key: str, default '!'
        The string to force an literal interpretation of a value.
        When the string is isolated, a literal interpreation is assumed and the surrounding
        quotes are replaced.
        
        E.g. Suppose a:b:c produces a dictionary {'d': 1}.
        - 'MRSM{a:b:c}'    => {'d': 1}        : isolated
        - ' MRSM{a:b:c} '  => ' "{\'d\': 1}"' : not isolated
        - ' MRSM{!a:b:c} ' => ' {"d": 1}'     : literal

    keep_symlinks :
        If True, include the symlinks under the top-level key '_symlinks' (never written to a file).
        Defaults to True.
        
        Example:

        ```
        MRSM{meerschaum:connectors:main:host} => cf['meerschaum']['connectors']['main']['host']
        ``` 
    """

    _links = []
    def _find_symlinks(d, _keys: Optional[List[str]] = None):
        if _keys is None:
            _keys = []
        if not isinstance(d, dict):
            return
        for k, v in d.items():
            if isinstance(v, dict):
                _find_symlinks(v, _keys + [k])
            elif (leading_key + begin_key) in str(v):
                _links.append((_keys + [k], v))

    _find_symlinks(config)

    import json
    needle = leading_key + begin_key
    haystack = json.dumps(config, separators=(',', ':'))
    mod_haystack = list(str(haystack))
    buff = str(needle)
    max_index = len(haystack) - len(buff)

    patterns = {}
    isolated_patterns = {}
    literal_patterns = {}

    begin, end, floor = 0, 0, 0
    while needle in haystack[floor:]:
        ### extract the keys
        hs = haystack[floor:]

        ### the first character of the keys
        ### MRSM{key1:key2}
        ###      ^
        begin = hs.find(needle) + len(needle)

        ### The character behind the needle.
        ### "MRSM{key1:key2}"
        ### ^
        prior = haystack[(floor + begin) - (len(needle) + 1)]

        ### number of characters to end of keys
        ### (really it's the index of the beginning of the end_key relative to the beginning
        ###     but the math works out)
        ### MRSM{key1}
        ###      ^   ^  => 4
        length = hs[begin:].find(end_key)

        ### index of the end_key (end of `length` characters)
        end = begin + length

        ### The character after the end_key.
        after = haystack[floor + end + 1]

        ### advance the floor to find the next leading key
        floor += end + len(end_key)
        pattern_keys = hs[begin:end].split(delimiter)

        ### Check for isolation key and empty keys (MRSM{}).
        force_literal = False
        keys = [k for k in pattern_keys]
        if str(keys[0]).startswith(literal_key):
            keys[0] = str(keys[0])[len(literal_key):]
            force_literal = True
        if len(keys) == 1 and keys[0] == '':
            keys = []

        ### Evaluate the parsed keys to extract the referenced value.
        ### TODO This needs to be recursive for chaining symlinks together.
        try:
            valid, value = get_config(
                *keys,
                substitute = False,
                as_tuple = True,
                write_missing = False,
                sync_files = False,
            )
        except Exception as e:
            import traceback
            traceback.print_exc()
            valid = False
        if not valid:
            continue

        ### pattern to search and replace
        pattern = leading_key + begin_key + delimiter.join(pattern_keys) + end_key

        ### store patterns and values
        patterns[pattern] = value

        ### Determine whether the pattern occured inside a string or is an isolated, direct symlink.
        isolated_patterns[pattern] = (prior == '"' and after == '"')

        literal_patterns[pattern] = force_literal

    ### replace the patterns with the values
    for pattern, value in patterns.items():
        if isolated_patterns[pattern]:
            haystack = haystack.replace(
                json.dumps(pattern),
                json.dumps(value),
            )
        elif literal_patterns[pattern]:
            haystack = haystack.replace(
                pattern,
                (
                    json.dumps(value)
                    .replace("\\", "\\\\")
                    .replace('"', '\\"')
                    .replace("'", "\\'")
                )
            )
        else:
            haystack = haystack.replace(pattern, str(value))

    ### parse back into dict
    parsed_config = json.loads(haystack) or {}

    symlinks = {}
    if keep_symlinks:
        ### Keep track of symlinks for writing back to a file.
        for _keys, _pattern in _links:
            s = symlinks
            for k in _keys[:-1]:
                if k not in s:
                    s[k] = {}
                s = s[k]
            s[_keys[-1]] = _pattern

        from meerschaum.config._patch import apply_patch_to_config
        from meerschaum.config.static import STATIC_CONFIG
        symlinks_key = STATIC_CONFIG['config']['symlinks_key']
        if symlinks_key not in parsed_config:
            parsed_config[symlinks_key] = {}
        parsed_config[symlinks_key] = apply_patch_to_config(parsed_config[symlinks_key], symlinks)

    return parsed_config


def get_possible_keys() -> List[str]:
    """
    Return a list of possible top-level keys.
    """
    import os
    from meerschaum.config._paths import CONFIG_DIR_PATH
    from meerschaum.config._default import default_config
    keys = set()
    for key in default_config:
        keys.add(key)
    for filename in os.listdir(CONFIG_DIR_PATH):
        keys.add('.'.join(filename.split('.')[:-1]))
    return sorted(list(keys))


def get_keyfile_path(
        key: str,
        create_new: bool = False,
        directory: Union[pathlib.Path, str, None] = None,
    ) -> Union[pathlib.Path, None]:
    """Determine a key's file path."""
    import os, pathlib
    if directory is None:
        from meerschaum.config._paths import CONFIG_DIR_PATH
        directory = CONFIG_DIR_PATH

    try:
        return pathlib.Path(
            os.path.join(
                directory,
                read_config(
                    keys = [key],
                    with_filenames = True,
                    write_missing = False,
                    substitute = False,
                )[1][0]
            )
        )
    except IndexError as e:
        if create_new:
            from meerschaum.config.static import STATIC_CONFIG
            default_filetype = STATIC_CONFIG['config']['default_filetype']
            return pathlib.Path(os.path.join(directory, key + '.' + default_filetype))
        return None



#! /usr/bin/env python3
# -*- coding: utf-8 -*-
# vim:fenc=utf-8

"""
Interactive "wizards" to guide the user.
"""

from __future__ import annotations

def select_pipes(
        yes: bool = False,
        force: bool = False,
        debug: bool = False,
    ) -> List[Pipe]:
    """Prompt the user for the keys to identify a list of pipes."""
    from meerschaum.utils.misc import get_connector_labels
    from meerschaum.utils.prompt import prompt, choose, yes_no
    from meerschaum.utils import get_pipes
    from meerschaum.utils.formatting._shell import clear_screen
    from meerschaum.utils.formatting import pprint_pipes
    from meerschaum.config import get_config
    from meerschaum.utils.warnings import warn
    from meerschaum.utils.misc import flatten_pipes_dict
    from meerschaum.connectors import instance_types
    clear_screen(debug=debug)
    while True:
        instance = choose(
            "On which instance are the pipes stored?",
            get_connector_labels(*instance_types),
            numeric = True,
            default = get_config('meerschaum', 'instance'),
        )
        pipes_dict = get_pipes(instance=instance, debug=debug)
        if pipes_dict:
            break
        warn(f"There are no pipes registered on instance '{instance}'.", stack=False)
        print("    Please choose another instance.")

    conn_keys = sorted(list(pipes_dict.keys()))
    clear_screen(debug=debug)
    chosen_conn_keys = choose(
        "What are the connectors for the pipes?",
        conn_keys,
        numeric = True,
        multiple = True,
        default = (conn_keys[0] if len(conn_keys) == 1 else None)
    )

    pipes_dict = get_pipes(chosen_conn_keys, instance=instance, debug=debug)
    metric_keys = []
    for ck, metrics in pipes_dict.items():
        for mk in metrics:
            metric_keys.append(mk)
    metric_keys = sorted(metric_keys)
    clear_screen(debug=debug)
    chosen_metric_keys = choose(
        "What are the metrics for the pipes?",
        metric_keys,
        numeric = True,
        multiple = True,
        default = (metric_keys[0] if len(metric_keys) == 1 else None)
    )
   
    pipes_dict = get_pipes(chosen_conn_keys, chosen_metric_keys, instance=instance, debug=debug)
    location_keys = []
    for ck, metrics in pipes_dict.items():
        for mk, locations in metrics.items():
            for lk, p in locations.items():
                location_keys.append(str(lk))
    location_keys = sorted(location_keys)
    clear_screen(debug=debug)
    chosen_location_keys = [(lk if lk != 'None' else None) for lk in choose(
        "What are the locations for the pipes?",
        location_keys,
        numeric = True,
        multiple = True,
        default = (location_keys[0] if len(location_keys) == 1 else None)
    )]

    pipes_dict = get_pipes(
        chosen_conn_keys,
        chosen_metric_keys,
        chosen_location_keys,
        instance = instance,
        debug = debug,
    )
    clear_screen(debug=debug)
    pprint_pipes(pipes_dict)
    if force or yes_no("Choose these pipes?", yes=yes):
        return flatten_pipes_dict(pipes_dict)
    return select_pipes(yes=yes, force=force, debug=debug)




#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:fenc=utf-8

"""
Copyright 2023 Bennett Meares

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import atexit
from meerschaum.utils.typing import SuccessTuple
from meerschaum.core.Pipe import Pipe
from meerschaum.plugins import Plugin
from meerschaum.utils.venv import Venv
from meerschaum.connectors import get_connector
from meerschaum.utils import get_pipes
from meerschaum.utils.formatting import pprint
from meerschaum._internal.docs import index as __doc__
from meerschaum.config import __version__, get_config
from meerschaum.utils.packages import attempt_import
from meerschaum.__main__ import _close_pools

atexit.register(_close_pools)

__pdoc__ = {'gui': False, 'api': False, 'core': False, '_internal': False}
__all__ = (
    "get_pipes",
    "get_connector",
    "get_config",
    "Pipe",
    "Plugin",
    "Venv",
    "Plugin",
    "pprint",
    "attempt_import",
    "actions",
    "config",
    "connectors",
    "plugins",
    "utils",
)



#! /usr/bin/env python3
# -*- coding: utf-8 -*-
# vim:fenc=utf-8

"""
Default configuration for jobs.
"""

default_jobs_config = {
    'timeout_seconds': 8,
    'check_timeout_interval_seconds': 0.1,
    'logs' : {
        'num_files_to_keep': 5,
        'max_file_size': 100_000,
        'lines_to_show': 30,
        'refresh_files_seconds': 5.0,
        'min_buffer_len': 15,
        'colors' : [
            'cyan',
            'magenta',
            'orange3',
            'green',
            'blue',
            'red',
            'spring_green3',
            'medium_purple3',
            'medium_violet_red',
            'slate_blue1', 
            'bright_red',
            'steel_blue3',
            'aquamarine1',
            'dark_khaki',
            'pink3',
            'gold3',
            'pale_green1',
            'light coral',
            'light_goldenrod2',
            'cornsilk1',
            'orange_red1',
            'deep_pink1',
            'aquamarine3',
            'sky_blue2',
            'tan',
            'honeydew2',
        ],
    },
}



#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:fenc=utf-8

"""
Edit a Pipe's parameters here.
"""

from __future__ import annotations
from meerschaum.utils.typing import Any, SuccessTuple

def update(self, *args, **kw) -> SuccessTuple:
    """
    Update a pipe's parameters in its instance.
    """
    kw['interactive'] = False
    return self.edit(*args, **kw)


def edit(
        self,
        patch: bool = False,
        interactive: bool = False,
        debug: bool = False,
        **kw: Any
    ) -> SuccessTuple:
    """
    Edit a Pipe's configuration.

    Parameters
    ----------
    patch: bool, default False
        If `patch` is True, update parameters by cascading rather than overwriting.
    interactive: bool, default False
        If `True`, open an editor for the user to make changes to the pipe's YAML file.
    debug: bool, default False
        Verbosity toggle.

    Returns
    -------
    A `SuccessTuple` of success, message.

    """
    from meerschaum.utils.venv import Venv
    from meerschaum.connectors import get_connector_plugin

    if self.temporary:
        return False, "Cannot edit pipes created with `temporary=True` (read-only)."

    if not interactive:
        with Venv(get_connector_plugin(self.instance_connector)):
            return self.instance_connector.edit_pipe(self, patch=patch, debug=debug, **kw)
    from meerschaum.config._paths import PIPES_CACHE_RESOURCES_PATH
    from meerschaum.utils.misc import edit_file
    parameters_filename = str(self) + '.yaml'
    parameters_path = PIPES_CACHE_RESOURCES_PATH / parameters_filename
    
    from meerschaum.utils.yaml import yaml

    edit_text = f"Edit the parameters for {self}"
    edit_top = '#' * (len(edit_text) + 4)
    edit_header = edit_top + f'\n# {edit_text} #\n' + edit_top + '\n\n'

    from meerschaum.config import get_config
    parameters = dict(get_config('pipes', 'parameters', patch=True))
    from meerschaum.config._patch import apply_patch_to_config
    parameters = apply_patch_to_config(parameters, self.parameters)

    ### write parameters to yaml file
    with open(parameters_path, 'w+') as f:
        f.write(edit_header)
        yaml.dump(parameters, stream=f, sort_keys=False)

    ### only quit editing if yaml is valid
    editing = True
    while editing:
        edit_file(parameters_path)
        try:
            with open(parameters_path, 'r') as f:
                file_parameters = yaml.load(f.read())
        except Exception as e:
            from meerschaum.utils.warnings import warn
            warn(f"Invalid format defined for '{self}':\n\n{e}")
            input(f"Press [Enter] to correct the configuration for '{self}': ")
        else:
            editing = False

    self.parameters = file_parameters

    if debug:
        from meerschaum.utils.formatting import pprint
        pprint(self.parameters)

    with Venv(get_connector_plugin(self.instance_connector)):
        return self.instance_connector.edit_pipe(self, patch=patch, debug=debug, **kw)


def edit_definition(
        self,
        yes: bool = False,
        noask: bool = False,
        force: bool = False,
        debug : bool = False,
        **kw : Any
    ) -> SuccessTuple:
    """
    Edit a pipe's definition file and update its configuration.
    **NOTE:** This function is interactive and should not be used in automated scripts!

    Returns
    -------
    A `SuccessTuple` of success, message.

    """
    if self.temporary:
        return False, "Cannot edit pipes created with `temporary=True` (read-only)."

    from meerschaum.connectors import instance_types
    if (self.connector is None) or self.connector.type not in instance_types:
        return self.edit(interactive=True, debug=debug, **kw)

    import json
    from meerschaum.utils.warnings import info, warn
    from meerschaum.utils.debug import dprint
    from meerschaum.config._patch import apply_patch_to_config
    from meerschaum.utils.misc import edit_file

    _parameters = self.parameters
    if 'fetch' not in _parameters:
        _parameters['fetch'] = {}

    def _edit_api():
        from meerschaum.utils.prompt import prompt, yes_no
        info(
            f"Please enter the keys of the source pipe from '{self.connector}'.\n" +
            "Type 'None' for None, or empty when there is no default. Press [CTRL+C] to skip."
        )

        _keys = { 'connector_keys' : None, 'metric_key' : None, 'location_key' : None }
        for k in _keys:
            _keys[k] = _parameters['fetch'].get(k, None)

        for k, v in _keys.items():
            try:
                _keys[k] = prompt(k.capitalize().replace('_', ' ') + ':', icon=True, default=v)
            except KeyboardInterrupt:
                continue
            if _keys[k] in ('', 'None', '\'None\'', '[None]'):
                _keys[k] = None

        _parameters['fetch'] = apply_patch_to_config(_parameters['fetch'], _keys)

        info("You may optionally specify additional filter parameters as JSON.")
        print("  Parameters are translated into a 'WHERE x AND y' clause, and lists are IN clauses.")
        print("  For example, the following JSON would correspond to 'WHERE x = 1 AND y IN (2, 3)':")
        print(json.dumps({'x': 1, 'y': [2, 3]}, indent=2, separators=(',', ': ')))
        if force or yes_no(
            "Would you like to add additional filter parameters?",
            yes=yes, noask=noask
        ):
            from meerschaum.config._paths import PIPES_CACHE_RESOURCES_PATH
            definition_filename = str(self) + '.json'
            definition_path = PIPES_CACHE_RESOURCES_PATH / definition_filename
            try:
                definition_path.touch()
                with open(definition_path, 'w+') as f:
                    json.dump(_parameters.get('fetch', {}).get('params', {}), f, indent=2)
            except Exception as e:
                return False, f"Failed writing file '{definition_path}':\n" + str(e)

            _params = None
            while True:
                edit_file(definition_path)
                try:
                    with open(definition_path, 'r') as f:
                        _params = json.load(f)
                except Exception as e:
                    warn(f'Failed to read parameters JSON:\n{e}', stack=False)
                    if force or yes_no(
                        "Would you like to try again?\n  "
                        + "If not, the parameters JSON file will be ignored.",
                        noask=noask, yes=yes
                    ):
                        continue
                    _params = None
                break
            if _params is not None:
                if 'fetch' not in _parameters:
                    _parameters['fetch'] = {}
                _parameters['fetch']['params'] = _params

        self.parameters = _parameters
        return True, "Success"

    def _edit_sql():
        import pathlib, os, textwrap
        from meerschaum.config._paths import PIPES_CACHE_RESOURCES_PATH
        from meerschaum.utils.misc import edit_file
        definition_filename = str(self) + '.sql'
        definition_path = PIPES_CACHE_RESOURCES_PATH / definition_filename

        sql_definition = _parameters['fetch'].get('definition', None)
        if sql_definition is None:
            sql_definition = ''
        sql_definition = textwrap.dedent(sql_definition).lstrip()

        try:
            definition_path.touch()
            with open(definition_path, 'w+') as f:
                f.write(sql_definition)
        except Exception as e:
            return False, f"Failed writing file '{definition_path}':\n" + str(e)

        edit_file(definition_path)
        try:
            with open(definition_path, 'r') as f:
                file_definition = f.read()
        except Exception as e:
            return False, f"Failed reading file '{definition_path}':\n" + str(e)

        if sql_definition == file_definition:
            return False, f"No changes made to definition for {self}."

        if ' ' not in file_definition:
            return False, f"Invalid SQL definition for {self}."

        if debug:
            dprint("Read SQL definition:\n\n" + file_definition)
        _parameters['fetch']['definition'] = file_definition
        self.parameters = _parameters
        return True, "Success"

    locals()['_edit_' + str(self.connector.type)]()
    return self.edit(interactive=False, debug=debug, **kw)



#! /usr/bin/env python3
# -*- coding: utf-8 -*-
# vim:fenc=utf-8

"""
Manage virtual environments.
"""

from __future__ import annotations

from meerschaum.utils.typing import Optional, Union, Dict, List, Tuple
from meerschaum.utils.threading import RLock, get_ident

__all__ = sorted([
    'activate_venv', 'deactivate_venv', 'init_venv',
    'inside_venv', 'is_venv_active', 'venv_exec',
    'venv_executable', 'venv_exists', 'venv_target_path',
    'Venv', 'get_venvs', 'verify_venv',
])
__pdoc__ = {'Venv': True}

LOCKS = {
    'sys.path': RLock(),
    'active_venvs': RLock(),
    'venvs_active_counts': RLock(),
}

active_venvs = set()
active_venvs_counts: Dict[str, int] = {}
active_venvs_order: List[Optional[str]] = []
threads_active_venvs: Dict[int, 'set[str]'] = {}


def activate_venv(
        venv: Optional[str] = 'mrsm',
        color: bool = True,
        force: bool = False,
        debug: bool = False,
        **kw
    ) -> bool:
    """
    Create a virtual environment (if it doesn't exist) and add it to `sys.path` if necessary.

    Parameters
    ----------
    venv: Optional[str], default 'mrsm'
        The virtual environment to activate.

    color: bool, default True
        If `True`, include color in debug text.

    force: bool, default False
        If `True`, do not exit early even if the venv is currently active.

    debug: bool, default False
        Verbosity toggle.

    Returns
    -------
    A bool indicating whether the virtual environment was successfully activated.

    """
    thread_id = get_ident()
    if active_venvs_order and active_venvs_order[0] == venv:
        if not force:
            return True
    import sys, platform, os
    from meerschaum.config._paths import VIRTENV_RESOURCES_PATH
    if debug:
        from meerschaum.utils.debug import dprint
    if venv is not None:
        init_venv(venv=venv, debug=debug)
    with LOCKS['active_venvs']:
        if thread_id not in threads_active_venvs:
            threads_active_venvs[thread_id] = {}
        active_venvs.add(venv)
        if venv not in threads_active_venvs[thread_id]:
            threads_active_venvs[thread_id][venv] = 1
        else:
            threads_active_venvs[thread_id][venv] += 1

        target = str(venv_target_path(venv, debug=debug))
        if venv in active_venvs_order:
            sys.path.remove(target)
            try:
                active_venvs_order.remove(venv)
            except Exception as e:
                pass
        if venv is not None:
            sys.path.insert(0, target)
        else:
            if sys.path and sys.path[0] in (os.getcwd(), ''):
                sys.path.insert(1, target)
            else:
                sys.path.insert(0, target)
        try:
            active_venvs_order.insert(0, venv)
        except Exception as e:
            pass

    return True


def deactivate_venv(
        venv: str = 'mrsm',
        color: bool = True,
        debug: bool = False,
        previously_active_venvs: Union['set[str]', List[str], None] = None,
        force: bool = False,
        **kw
    ) -> bool:
    """
    Remove a virtual environment from `sys.path` (if it's been activated).

    Parameters
    ----------
    venv: str, default 'mrsm'
        The virtual environment to deactivate.

    color: bool, default True
        If `True`, include color in debug text.

    debug: bool, default False
        Verbosity toggle.

    previously_active_venvs: Union[Set[str], List[str], None]
        If provided, skip deactivating if a virtual environment is in this iterable.

    force: bool, default False
        If `True`, forcibly deactivate the virtual environment.
        This may cause issues with other threads, so be careful!

    Returns
    -------
    Return a bool indicating whether the virtual environment was successfully deactivated.

    """
    import sys
    thread_id = get_ident()
    if venv is None:
        if venv in active_venvs:
            active_venvs.remove(venv)
        return True

    if previously_active_venvs and venv in previously_active_venvs and not force:
        return True

    with LOCKS['active_venvs']:
        if venv in threads_active_venvs.get(thread_id, {}):
            new_count = threads_active_venvs[thread_id][venv] - 1
            if new_count > 0 and not force:
                threads_active_venvs[thread_id][venv] = new_count
                return True
            else:
                del threads_active_venvs[thread_id][venv]

        if not force:
            for other_thread_id, other_venvs in threads_active_venvs.items():
                if other_thread_id == thread_id:
                    continue
                if venv in other_venvs:
                    return True
        else:
            to_delete = [other_thread_id for other_thread_id in threads_active_venvs]
            for other_thread_id in to_delete:
                del threads_active_venvs[other_thread_id]

        if venv in active_venvs:
            active_venvs.remove(venv)

    if sys.path is None:
        return False

    target = str(venv_target_path(venv, allow_nonexistent=force, debug=debug))
    with LOCKS['sys.path']:
        if target in sys.path:
            sys.path.remove(target)
            try:
                active_venvs_order.remove(venv)
            except Exception as e:
                pass

    return True


def is_venv_active(
        venv: str = 'mrsm',
        color : bool = True,
        debug: bool = False
    ) -> bool:
    """
    Check if a virtual environment is active.

    Parameters
    ----------
    venv: str, default 'mrsm'
        The virtual environment to check.

    color: bool, default True
        If `True`, include color in debug text.

    debug: bool, default False
        Verbosity toggle.

    Returns
    -------
    A bool indicating whether the virtual environment `venv` is active.

    """
    return venv in active_venvs


verified_venvs = set()
def verify_venv(
        venv: str,
        debug: bool = False,
    ) -> None:
    """
    Verify that the virtual environment matches the expected state.
    """
    import pathlib, platform, os, shutil, subprocess, sys
    from meerschaum.config._paths import VIRTENV_RESOURCES_PATH
    from meerschaum.utils.process import run_process
    from meerschaum.utils.misc import make_symlink, is_symlink
    from meerschaum.utils.warnings import warn
    venv_path = VIRTENV_RESOURCES_PATH / venv
    bin_path = venv_path / (
        'bin' if platform.system() != 'Windows' else "Scripts"
    )
    current_python_versioned_name = (
        'python' + str(sys.version_info.major) + '.' + str(sys.version_info.minor)
        + ('' if platform.system() != 'Windows' else '.exe')
    )

    if not (bin_path / current_python_versioned_name).exists():
        init_venv(venv, verify=False, force=True, debug=debug)
        current_python_in_venv_path = pathlib.Path(venv_executable(venv=venv))
        current_python_in_sys_path = pathlib.Path(venv_executable(venv=None))
        if not current_python_in_venv_path.exists():
            if is_symlink(current_python_in_venv_path):
                try:
                    current_python_in_venv_path.unlink()
                except Exception as e:
                    print(f"Unable to remove symlink {current_python_in_venv_path}:\n{e}")
            try:
                make_symlink(current_python_in_sys_path, current_python_in_venv_path)
            except Exception as e:
                print(
                    f"Unable to create symlink {current_python_in_venv_path} "
                    + f"to {current_python_in_sys_path}."
                )
        files_to_inspect = sorted(os.listdir(bin_path), reverse=True)
    else:
        files_to_inspect = [current_python_versioned_name]

    def get_python_version(python_path: pathlib.Path) -> Union[str, None]:
        """
        Return the version for the python binary at the given path.
        """
        try:
            ### It might be a broken symlink, so skip on errors.
            if debug:
                print(f"Getting python version for {python_path}")
            proc = run_process(
                [str(python_path), '-V'],
                as_proc = True,
                capture_output = True,
            )
            stdout, stderr = proc.communicate(timeout=1.0)
        except Exception as e:
            ### E.g. the symlink may be broken.
            if is_symlink(python_path):
                try:
                    python_path.unlink()
                except Exception as _e:
                    print(f"Unable to remove broken symlink {python_path}:\n{e}\n{_e}")
            return None
        return stdout.decode('utf-8').strip().replace('Python ', '')

    ### Ensure the versions are symlinked correctly.
    for filename in files_to_inspect:
        if not filename.startswith('python'):
            continue
        python_path = bin_path / filename
        version = get_python_version(python_path)
        if version is None:
            continue
        try:
            major_version = version.split('.', maxsplit=1)[0]
            minor_version = version.split('.', maxsplit=2)[1]
        except IndexError:
            return
        python_versioned_name = (
            'python' + major_version + '.' + minor_version
            + ('' if platform.system() != 'Windows' else '.exe')
        )

        ### E.g. python3.10 actually links to Python 3.10.
        if filename == python_versioned_name:
            real_path = pathlib.Path(os.path.realpath(python_path))
            if not real_path.exists():
                try:
                    python_path.unlink()
                except Exception as e:
                    pass
                init_venv(venv, verify=False, force=True, debug=debug)
                if not python_path.exists():
                    raise FileNotFoundError(f"Unable to verify Python symlink:\n{python_path}")

            if python_path == real_path:
                continue

            try:
                python_path.unlink()
            except Exception as e:
                pass
            success, msg = make_symlink(real_path, python_path)
            if not success:
                warn(msg, color=False)
            continue

        python_versioned_path = bin_path / python_versioned_name
        if python_versioned_path.exists():
            ### Avoid circular symlinks.
            if get_python_version(python_versioned_path) == version:
                continue
            python_versioned_path.unlink()
        shutil.move(python_path, python_versioned_path)


tried_virtualenv = False
def init_venv(
        venv: str = 'mrsm',
        verify: bool = True,
        force: bool = False,
        debug: bool = False,
    ) -> bool:
    """
    Initialize the virtual environment.

    Parameters
    ----------
    venv: str, default 'mrsm'
        The name of the virtual environment to create.

    verify: bool, default True
        If `True`, verify that the virtual environment is in the expected state.

    force: bool, default False
        If `True`, recreate the virtual environment, even if already initalized.

    Returns
    -------
    A `bool` indicating success.
    """
    if not force and venv in verified_venvs:
        return True
    if not force and venv_exists(venv, debug=debug):
        if verify:
            verify_venv(venv, debug=debug)
            verified_venvs.add(venv)
        return True

    import sys, platform, os, pathlib, shutil
    from meerschaum.config.static import STATIC_CONFIG
    from meerschaum.config._paths import VIRTENV_RESOURCES_PATH
    venv_path = VIRTENV_RESOURCES_PATH / venv
    docker_home_venv_path = pathlib.Path('/home/meerschaum/venvs/mrsm')

    runtime_env_var = STATIC_CONFIG['environment']['runtime']
    work_dir_env_var = STATIC_CONFIG['environment']['work_dir']
    if (
        not force
        and venv == 'mrsm'
        and os.environ.get(work_dir_env_var, None) is not None
        and docker_home_venv_path.exists()
    ):
        shutil.move(docker_home_venv_path, venv_path)
        if verify:
            verify_venv(venv, debug=debug)
            verified_venvs.add(venv)
        return True

    from meerschaum.utils.packages import run_python_package, attempt_import
    global tried_virtualenv
    try:
        import venv as _venv
        virtualenv = None
    except ImportError:
        _venv = None
        virtualenv = None
    

    _venv_success = False
    if _venv is not None:
        import io
        from contextlib import redirect_stdout
        f = io.StringIO()
        with redirect_stdout(f):
            _venv_success = run_python_package(
                'venv',
                [str(venv_path)] + (
                    ['--symlinks'] if platform.system() != 'Windows' else []
                ),
                venv=None, debug=debug
            ) == 0
        if not _venv_success:
            print(f"Please install python3-venv.\n{f.getvalue()}\nFalling back to virtualenv...")
        if not venv_exists(venv, debug=debug):
            _venv = None
    if not _venv_success:
        virtualenv = attempt_import(
            'virtualenv', venv=None, lazy=False, install=(not tried_virtualenv), warn=False,
            check_update=False, color=False, debug=debug,
        )
        if virtualenv is None:
            print(
                "Failed to import `venv` or `virtualenv`! "
                + "Please install `virtualenv` via pip then restart Meerschaum."
            )
            return False

        tried_virtualenv = True
        try:
            python_folder = (
                'python' + str(sys.version_info.major) + '.' + str(sys.version_info.minor)
            )
            dist_packages_path = (
                VIRTENV_RESOURCES_PATH /
                venv / 'local' / 'lib' / python_folder / 'dist-packages'
            )
            local_bin_path = VIRTENV_RESOURCES_PATH / venv / 'local' / 'bin'
            bin_path = VIRTENV_RESOURCES_PATH / venv / 'bin'
            vtp = venv_target_path(venv=venv, allow_nonexistent=True, debug=debug)
            if bin_path.exists():
                try:
                    shutil.rmtree(bin_path)
                except Exception as e:
                    import traceback
                    traceback.print_exc()
            virtualenv.cli_run([str(venv_path)])
            if dist_packages_path.exists():
                vtp.mkdir(exist_ok=True, parents=True)
                for file_path in dist_packages_path.glob('*'):
                    shutil.move(file_path, vtp)
                shutil.rmtree(dist_packages_path)
                #  shutil.move(dist_packages_path, vtp)
                bin_path.mkdir(exist_ok=True, parents=True)
                for file_path in local_bin_path.glob('*'):
                    shutil.move(file_path, bin_path)
                #  shutil.move(local_bin_path, bin_path)
                shutil.rmtree(local_bin_path)

        except Exception as e:
            import traceback
            traceback.print_exc()
            return False
    if verify:
        verify_venv(venv, debug=debug)
        verified_venvs.add(venv)
    return True


def venv_executable(venv: Optional[str] = 'mrsm') -> str:
    """
    The Python interpreter executable for a given virtual environment.
    """
    from meerschaum.config._paths import VIRTENV_RESOURCES_PATH
    import sys, platform, os
    return (
        sys.executable if venv is None
        else str(
            VIRTENV_RESOURCES_PATH
            / venv
            / (
                'bin' if platform.system() != 'Windows'
                else 'Scripts'
            ) / (
                'python'
                + str(sys.version_info.major)
                + '.'
                + str(sys.version_info.minor)
            )
        )
    )


def venv_exec(
        code: str,
        venv: Optional[str] = 'mrsm',
        env: Optional[Dict[str, str]] = None,
        with_extras: bool = False,
        as_proc: bool = False,
        capture_output: bool = True,
        debug: bool = False,
    ) -> Union[bool, Tuple[int, bytes, bytes]]:
    """
    Execute Python code in a subprocess via a virtual environment's interpeter.
    Return `True` if the code successfully executes, `False` on failure.

    Parameters
    ----------
    code: str
        The Python code to excecute.

    venv: str, default 'mrsm'
        The virtual environment to use to get the path for the Python executable.
        If `venv` is `None`, use the default `sys.executable` path.

    env: Optional[Dict[str, str]], default None
        Optionally specify the environment variables for the subprocess.
        Defaults to `os.environ`.

    with_extras: bool, default False
        If `True`, return a tuple of the exit code, stdout bytes, and stderr bytes.

    as_proc: bool, default False
        If `True`, return the `subprocess.Popen` object instead of executing.

    Returns
    -------
    By default, return a bool indicating success.
    If `as_proc` is `True`, return a `subprocess.Popen` object.
    If `with_extras` is `True`, return a tuple of the exit code, stdout bytes, and stderr bytes.

    """
    import os
    import subprocess
    from meerschaum.utils.debug import dprint
    executable = venv_executable(venv=venv)
    cmd_list = [executable, '-c', code]
    if env is None:
        env = os.environ
    if debug:
        dprint(str(cmd_list))
    if not with_extras and not as_proc:
        return subprocess.call(cmd_list, env=env) == 0

    stdout, stderr = (None, None) if not capture_output else (subprocess.PIPE, subprocess.PIPE)
    process = subprocess.Popen(
        cmd_list,
        stdout = stdout,
        stderr = stderr,
        env = env,
    )
    if as_proc:
        return process
    stdout, stderr = process.communicate()
    exit_code = process.returncode
    return exit_code, stdout, stderr


def venv_exists(venv: Union[str, None], debug: bool = False) -> bool:
    """
    Determine whether a virtual environment has been created.
    """
    target_path = venv_target_path(venv, allow_nonexistent=True, debug=debug)
    return target_path.exists()


def venv_target_path(
        venv: Union[str, None],
        allow_nonexistent: bool = False,
        debug: bool = False,
    ) -> 'pathlib.Path':
    """
    Return a virtual environment's site-package path.

    Parameters
    ----------
    venv: Union[str, None]
        The virtual environment for which a path should be returned.

    allow_nonexistent: bool, default False
        If `True`, return a path even if it does not exist.

    Returns
    -------
    The `pathlib.Path` object for the virtual environment's path.

    """
    import os, sys, platform, pathlib, site
    from meerschaum.config._paths import VIRTENV_RESOURCES_PATH
    from meerschaum.config.static import STATIC_CONFIG

    ### Check sys.path for a user-writable site-packages directory.
    if venv is None:

        ### Return the known value for the portable environment.
        environment_runtime = STATIC_CONFIG['environment']['runtime']
        if os.environ.get(environment_runtime, None) == 'portable':
            python_version_folder = (
                'python' + str(sys.version_info.major) + '.' + str(sys.version_info.minor)
            )
            executable_path = pathlib.Path(sys.executable)
            site_packages_path = (
                (
                    executable_path.parent.parent / 'lib' / python_version_folder / 'site-packages'
                ) if platform.system() != 'Windows' else (
                    executable_path.parent / 'Lib' / 'site-packages'
                )
            )
            if not site_packages_path.exists():
                raise EnvironmentError(f"Could not find '{site_packages_path}'. Does it exist?")
            return site_packages_path

        if not inside_venv():
            site_path = pathlib.Path(site.getusersitepackages())
            if not site_path.exists():

                ### Windows does not have `os.geteuid()`.
                if platform.system() == 'Windows' or os.geteuid() != 0:
                    site_path.mkdir(parents=True, exist_ok=True)
                    return site_path

                ### Allow for dist-level paths (running as root).
                for possible_dist in reversed(site.getsitepackages()):
                    dist_path = pathlib.Path(possible_dist)
                    if not dist_path.exists():
                        continue
                    return dist_path
                
                raise EnvironmentError("Could not determine the dist-packages directory.")

            return site_path

    venv_root_path = (
        (VIRTENV_RESOURCES_PATH / venv)
        if venv is not None else pathlib.Path(sys.prefix)
    )
    target_path = venv_root_path

    ### Ensure 'lib' or 'Lib' exists.
    lib = 'lib' if platform.system() != 'Windows' else 'Lib'
    if not allow_nonexistent:
        if not venv_root_path.exists() or lib not in os.listdir(venv_root_path):
            print(f"Failed to find lib directory for virtual environment '{venv}'.")
            import traceback
            traceback.print_stack()
            sys.exit(1)
    target_path = target_path / lib

    ### Check if a 'python3.x' folder exists.
    python_folder = 'python' + str(sys.version_info.major) + '.' + str(sys.version_info.minor)
    if target_path.exists():
        target_path = (
            (target_path / python_folder) if python_folder in os.listdir(target_path)
            else target_path
        )
    else:
        target_path = (
            (target_path / python_folder) if platform.system() != 'Windows'
            else target_path
        )

    ### Ensure 'site-packages' exists.
    if allow_nonexistent or 'site-packages' in os.listdir(target_path): ### Windows
        target_path = target_path / 'site-packages'
    else:
        import traceback
        traceback.print_stack()
        print(f"Failed to find site-packages directory for virtual environment '{venv}'.")
        print("This may be because you are using a different Python version.")
        print("Try deleting the following directory and restarting Meerschaum:")
        print(VIRTENV_RESOURCES_PATH)
        sys.exit(1)

    return target_path


def inside_venv() -> bool:
    """
    Determine whether current Python interpreter is running inside a virtual environment.
    """
    import sys
    return (
        hasattr(sys, 'real_prefix') or (
            hasattr(sys, 'base_prefix')
                and sys.base_prefix != sys.prefix
        )
    )


def get_venvs() -> List[str]:
    """
    Return a list of all the virtual environments.
    """
    import os
    from meerschaum.config._paths import VIRTENV_RESOURCES_PATH
    venvs = []
    for filename in os.listdir(VIRTENV_RESOURCES_PATH):
        path = VIRTENV_RESOURCES_PATH / filename
        if not path.is_dir():
            continue
        if not venv_exists(filename):
            continue
        venvs.append(filename)
    return venvs


from meerschaum.utils.venv._Venv import Venv



#! /usr/bin/env python3
# -*- coding: utf-8 -*-
# vim:fenc=utf-8

"""
Patch the runtime configuration from environment variables.
"""

import os
import re
import json
from meerschaum.utils.typing import List, Union, Dict, Any
from meerschaum.config.static import STATIC_CONFIG

def apply_environment_patches() -> None:
    """
    Apply patches defined in `MRSM_CONFIG` and `MRSM_PATCH`.
    """
    config_var = STATIC_CONFIG['environment']['config']
    patch_var = STATIC_CONFIG['environment']['patch']
    apply_environment_config(config_var)
    apply_environment_config(patch_var)


def apply_environment_config(env_var: str) -> None:
    """
    Parse a dictionary (simple or JSON) from an environment variable
    and apply it to the current configuration.
    """
    from meerschaum.config import get_config, set_config, _config
    from meerschaum.config._patch import apply_patch_to_config
    if env_var not in os.environ:
        return
    from meerschaum.utils.misc import string_to_dict
    try:
        _patch = string_to_dict(str(os.environ[env_var]).lstrip())
    except Exception as e:
        _patch = None
    error_msg = (
        f"Environment variable {env_var} is set but cannot be parsed.\n"
        f"Unset {env_var} or change to JSON or simplified dictionary format " +
        "(see --help, under params for formatting)\n" +
        f"{env_var} is set to:\n{os.environ[env_var]}\n"
        f"Skipping patching os environment into config..."
    )

    if not isinstance(_patch, dict):
        print(error_msg)
        return

    valids = []

    def load_key(key: str) -> Union[Dict[str, Any], None]:
        try:
            c = get_config(key, warn=False)
        except Exception as e:
            c = None
        return c

    ### This was multi-threaded, but I ran into all sorts of locking issues.
    keys = list(_patch.keys())
    for key in keys:
        _ = load_key(key)

    ### Load and patch config files.
    set_config(
        apply_patch_to_config(
            _config(),
            _patch,
        )
    )


def apply_environment_uris() -> None:
    """
    Patch temporary connectors defined in environment variables which start with
    `MRSM_SQL_` or `MRSM_API_`.
    """
    for env_var in get_connector_env_vars():
        apply_connector_uri(env_var)


def get_connector_env_regex() -> str:
    """
    Return the regex pattern for valid environment variable names for instance connectors.
    """
    return STATIC_CONFIG['environment']['uri_regex']


def get_connector_env_vars() -> List[str]:
    """
    Get the names of the environment variables which match the Meerschaum connector regex.

    Examples
    --------
    >>> get_connector_environment_vars()
    ['MRSM_SQL_FOO']
    """
    uri_regex = get_connector_env_regex()
    env_vars = []
    for env_var in os.environ:
        matched = re.match(uri_regex, env_var)
        if matched is None:
            continue
        if env_var in STATIC_CONFIG['environment'].values():
            continue
        env_vars.append(env_var)
    return env_vars


def apply_connector_uri(env_var: str) -> None:
    """
    Parse and validate a URI obtained from an environment variable.
    """
    from meerschaum.config import get_config, set_config, _config
    from meerschaum.config._patch import apply_patch_to_config
    from meerschaum.config._read_config import search_and_substitute_config
    from meerschaum.utils.warnings import warn
    uri_regex = get_connector_env_regex()
    matched = re.match(uri_regex, env_var)
    groups = matched.groups()
    typ, label = groups[0].lower(), groups[1].lower()
    uri = os.environ[env_var]
    if uri.lstrip().startswith('{') and uri.rstrip().endswith('}'):
        try:
            conn_attrs = json.loads(uri)
        except Exception as e:
            warn(f"Unable to parse JSON for environment connector '{typ}:{label}'.")
            conn_attrs = {'uri': uri}
    else:
        conn_attrs = {'uri': uri}
    set_config(
        apply_patch_to_config(
            {'meerschaum': get_config('meerschaum')},
            {'meerschaum': {'connectors': {typ: {label: conn_attrs}}}},
        )
    )


def get_env_vars() -> List[str]:
    """
    Return all environment variables which begin with `'MRSM_'`.
    """
    prefix = STATIC_CONFIG['environment']['prefix']
    return sorted([env_var for env_var in os.environ if env_var.startswith(prefix)])



#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:fenc=utf-8

"""
Execute actions via the web interface.
"""

from __future__ import annotations
import platform, sys, io, os, shlex, time, copy
from dash.exceptions import PreventUpdate
from meerschaum.utils.threading import Thread
from meerschaum.utils.typing import SuccessTuple, Tuple, Dict, Any, WebState
from meerschaum.utils.packages import attempt_import, import_html, import_dcc
from meerschaum.utils.misc import remove_ansi
from meerschaum.actions import actions, get_shell
from meerschaum.api import debug, CHECK_UPDATE
from meerschaum.api.dash import (
    running_jobs, stopped_jobs, running_monitors, stopped_monitors, active_sessions
)
from meerschaum.api import get_api_connector
from meerschaum.api.dash.connectors import get_web_connector
from meerschaum.api.dash.components import alert_from_success_tuple, console_div
from meerschaum.api.dash.webterm import get_webterm
from meerschaum.api.dash.pipes import pipes_from_state, keys_from_state
from meerschaum.api.dash.websockets import ws_send
from meerschaum.api._websockets import websockets
html, dcc = import_html(check_update=CHECK_UPDATE), import_dcc(check_update=CHECK_UPDATE)
from meerschaum.config import get_config
from meerschaum.core import User

def execute_action(state: WebState):
    """
    Execute a Meerschaum action and capture its output.
    Format the output as an HTML `pre` object, and return a list of Alert objects.

    Parameters
    ----------
    state: WebState :
        The state of the dashboard.

    Returns
    -------
    A tuple of lists.
    The first list contains the children of the output div.
    The second contains alerts.

    """
    global running_jobs, stopped_jobs, running_monitors, stopped_monitors
    action, subaction, subaction_options, subaction_hidden, additional_subaction_text, flags = (
        state['action-dropdown.value'],
        state['subaction-dropdown.value'],
        state['subaction-dropdown.options'],
        state['subaction-dropdown-div.hidden'],
        state['subaction-dropdown-text.value'],
        state['flags-dropdown.value'],
    )
    if action is None:
        return [], []
    if subaction is None or subaction_hidden:
        subaction = None
    ### If the current value is not a member of the options, choose the first element (visible).
    elif subaction not in [o['value'] for o in subaction_options]:
        subaction = subaction_options[0]['value']
    if additional_subaction_text is None:
        additional_subaction_text = ''
    if flags is None:
        flags = []

    ### TODO add direct Input to parse if active_tab is 'text'

    session_id = state['session-store.data'].get('session-id', None)

    ### Check if actions are permitted by non-admin users.
    permissions = get_config('system', 'api', 'permissions')
    allow_non_admin = permissions.get('actions', {}).get('non_admin', False)
    if not allow_non_admin:
        username = active_sessions.get(session_id, {}).get('username', None)
        user = User(username, instance=get_api_connector())
        user_type = get_api_connector().get_user_type(user, debug=debug)
        if user_type != 'admin':
            msg = (
                f"User '{username}' is not authorized to perform actions on this instance.\n\n"
                + "To allow non-administrator users to execute actions,\n"
                + "  run the command `mrsm edit config system`.\n\n"
                + "Under the keys `api:permissions:actions`, set the value of \n"
                + "  `non_admin` to `true`."
            )
            return (
                [
                    html.Div(
                        html.Pre(msg, id='console-pre'),
                        id = 'console-div'
                    ),
                ],
                [alert_from_success_tuple(
                    (False, "Actions are not allowed on this Meerschaum instance.")
                )]
            )

    
    subactions = ([subaction] if subaction else []) + shlex.split(additional_subaction_text)

    keywords = {f : True for f in flags}
    keywords['debug'] = keywords.get('debug', debug)
    _ck, _mk, _lk, _params = keys_from_state(state, with_params=True)
    keywords['connector_keys'], keywords['metric_keys'], keywords['location_keys'] = (
        _ck, _mk, _lk,
    )
    keywords['params'] = _params
    keywords['mrsm_instance'] = get_web_connector(state) 

    def do_action() -> SuccessTuple:
        if sys.stdout is not cap:
            stdout = sys.stdout
            sys.stdout = cap
        try:
            success_tuple = actions[action](action=subactions, **keywords)
        except Exception as e:
            success_tuple = False, str(e)
        except KeyboardInterrupt:
            success_tuple = False, f"Action '{action}' was manually cancelled."
        stopped_jobs[session_id] = running_jobs.pop(session_id)
        stopped_monitors[session_id] = running_monitors.pop(session_id)
        return success_tuple

    cap = io.StringIO()
    _sentinel = object()
    def monitor_and_send_stdout():
        cap_buffer = ''
        while True:
            if not (sys.stdout) is cap:
                sys.stdout = cap
            if session_id not in running_jobs:
                text = cap.getvalue()
                print('Stopping monitoring and sending final message.', file=sys.stderr)
                ws_send(text, session_id)
                break
            text = cap.getvalue()
            if not text or len(text) == len(cap_buffer):
                continue
            cap_buffer = text
            ws_send(text, session_id)
            ### So we don't overwhelm the client.
            time.sleep(0.01)

    def use_stringio():
        try:
            LINES, COLUMNS = (
                os.environ.get('LINES', str(os.get_terminal_size().lines)),
                os.environ.get('COLUMNS', str(os.get_terminal_size().columns)),
            )
        except OSError:
            LINES, COLUMNS = '120', '100'
        os.environ['LINES'], os.environ['COLUMNS'] = '120', '100'
        stdout = sys.stdout
        success_tuple = use_process()
        os.environ['LINES'], os.environ['COLUMNS'] = LINES, COLUMNS
        text = cap.getvalue()
        return text, success_tuple


    def use_process():
        from meerschaum.utils.packages import run_python_package
        from meerschaum._internal.arguments._parse_arguments import parse_dict_to_sysargs
        max_buffer_size = 10000
        line_buffer = ''
        def send_line(line: str):
            nonlocal line_buffer
            line_buffer += line
            buff_start_idx = max(len(line_buffer) - max_buffer_size, 0)
            line_buffer = line_buffer[buff_start_idx:]
            ws_send(line_buffer, session_id)

        def do_process():
            keywords['action'] = [action] + subactions
            _sysargs = parse_dict_to_sysargs(keywords)

            ### Don't forget to pass the existing environment
            ### in case MRSM_CONFIG and MRSM_PATCH are set.
            try:
                _env = copy.deepcopy(os.environ)
            except Exception as e:
                _env = {}
            _env.update({'LINES': '120', 'COLUMNS': '100'})

            process = run_python_package(
                'meerschaum', _sysargs,
                line_callback = send_line,
                env = _env,
                foreground = False,
                universal_newlines = True,
                as_proc = True,
                capture_output = False,
                ### Store the `subprocess.Popen` process in case we need to terminate it later.
                store_proc_dict = running_jobs[session_id],
                venv = None,
                debug = debug,
            )
        action_thread = Thread(target=do_process)
        running_jobs[session_id] = {'parent_thread': action_thread,}
        action_thread.start()
        return True, "Success"


    def use_thread():
        action_thread = Thread(target=do_action, daemon=True)
        monitor_thread = Thread(target=monitor_and_send_stdout) 
        running_jobs[session_id] = action_thread
        running_monitors[session_id] = monitor_thread
        monitor_thread.start()
        action_thread.start()
        success_tuple = True, 'Success'
        return success_tuple

    text, success_tuple = use_stringio()

    return (
        [
            html.Div(
                html.Pre(text, id='console-pre'),
                id = 'console-div'
            ),
        ],
        [alert_from_success_tuple(success_tuple)]
    )


def stop_action(state: WebState):
    """
    Kill a running action.

    Parameters
    ----------
    state: WebState:
        The state of the dashboard.

    Returns
    -------
    A tuple of children and alerts.
    """
    global running_jobs
    session_id = state['session-store.data'].get('session-id', None)
    thread = running_jobs.get(session_id, {}).get('parent_thread', None)
    proc = running_jobs.get(session_id, {}).get('child_process', None)
    if proc is not None and proc.poll() is None:
        proc.terminate()
        thread.join()
    success_tuple = True, "Success"
    webterm_iframe, webterm_alerts = get_webterm(state)
    console_to_return = console_div if webterm_alerts else webterm_iframe
    return (
        [console_to_return],
        [alert_from_success_tuple(success_tuple)]
    )



#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:fenc=utf-8

"""
Declare FastAPI events in this module (startup, shutdown, etc.).
"""

import sys, os, time
from meerschaum.api import (
    app,
    get_api_connector,
    get_uvicorn_config,
    debug,
    no_dash,
    uvicorn_config_path,
)
from meerschaum.utils.debug import dprint
from meerschaum.connectors.poll import retry_connect
from meerschaum.utils.warnings import warn
from meerschaum._internal.term.tools import is_webterm_running

@app.on_event("startup")
async def startup():
    conn = get_api_connector()
    try:
        if not no_dash:
            from meerschaum.api.dash.webterm import start_webterm
            start_webterm()
        connected = retry_connect(
            get_api_connector(),
            workers = get_uvicorn_config().get('workers', None),
            debug = debug
        )
    except Exception as e:
        import traceback
        traceback.print_exc()
        connected = False
    if not connected:
        await shutdown()
        os._exit(1)


@app.on_event("shutdown")
async def shutdown():
    if debug:
        dprint("Closing connection...")
    if get_api_connector().type == 'sql':
        get_api_connector().engine.dispose()

    ### Terminate any running jobs left over.
    if 'meerschaum.api.dash' in sys.modules:
        from meerschaum.api.dash.actions import running_jobs, stop_action
        from meerschaum.api.dash.webterm import stop_webterm
        stop_webterm()
        for session_id in running_jobs:
            stop_action({'session-store.data': {'session-id': session_id}})



#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:fenc=utf-8

"""
Callbacks for the main dashboard.
"""

from __future__ import annotations
import sys, textwrap, json, datetime, uuid
from dash.dependencies import Input, Output, State, ALL, MATCH
from dash.exceptions import PreventUpdate
from meerschaum.config import get_config
from meerschaum.config.static import _static_config
from meerschaum.utils.typing import List, Optional, Any, Tuple
from meerschaum.api import get_api_connector, endpoints, no_auth, CHECK_UPDATE
from meerschaum.api.dash import (
    dash_app, debug, pipes, _get_pipes,
    active_sessions, authenticated_sessions, unauthenticated_sessions,
)
from meerschaum.api.dash.users import is_session_authenticated
from meerschaum.api.dash.connectors import get_web_connector
from meerschaum.connectors.parse import parse_instance_keys
from meerschaum.api.dash.pipes import get_pipes_cards, pipe_from_ctx, accordion_items_from_pipe
from meerschaum.api.dash.jobs import get_jobs_cards
from meerschaum.api.dash.plugins import get_plugins_cards
from meerschaum.api.dash.users import get_users_cards
from meerschaum.api.dash.graphs import get_graphs_cards
from meerschaum.api.dash.webterm import get_webterm
from meerschaum.api.dash.components import (
    alert_from_success_tuple, console_div, build_cards_grid,
)
from meerschaum.api.dash.actions import execute_action, stop_action
import meerschaum.api.dash.pages as pages
from meerschaum.utils.typing import Dict
from meerschaum.utils.debug import dprint
from meerschaum.utils.packages import attempt_import, import_html, import_dcc
from meerschaum.utils.misc import (
    string_to_dict, get_connector_labels, json_serialize_datetime, filter_keywords,
    flatten_list,
)
from meerschaum.utils.yaml import yaml
from meerschaum.actions import get_subactions, actions
from meerschaum._internal.arguments._parser import get_arguments_triggers, parser
from meerschaum.connectors.sql._fetch import set_pipe_query
import meerschaum as mrsm
import json
dash = attempt_import('dash', lazy=False, check_update=CHECK_UPDATE)
dbc = attempt_import('dash_bootstrap_components', lazy=False, check_update=CHECK_UPDATE)
dcc, html = import_dcc(check_update=CHECK_UPDATE), import_html(check_update=CHECK_UPDATE)
dex = attempt_import('dash_extensions', lazy=False, check_update=CHECK_UPDATE)

keys_state = (
    State('connector-keys-dropdown', 'value'),
    State('metric-keys-dropdown', 'value'),
    State('location-keys-dropdown', 'value'),
    State('connector-keys-input', 'value'),
    State('metric-keys-input', 'value'),
    State('location-keys-input', 'value'),
    State('search-parameters-editor', 'value'),
    State('pipes-filter-tabs', 'active_tab'),
    State('action-dropdown', 'value'),
    State('subaction-dropdown', 'value'),
    State('subaction-dropdown', 'options'),
    State('subaction-dropdown-div', 'hidden'),
    State('subaction-dropdown-text', 'value'),
    State('flags-dropdown', 'value'),
    State({'type': 'input-flags-dropdown', 'index': ALL}, 'value'),
    State({'type': 'input-flags-dropdown-text', 'index': ALL}, 'value'),
    State('instance-select', 'value'),
    State('content-div-right', 'children'),
    State('success-alert-div', 'children'),
    State('session-store', 'data'),
)

omit_flags = {
    'help',
    'gui',
    'version',
    'shell',
    'use_bash',
    'trace',
    'allow_shell_job',
    'action',
    'mrsm_instance',
}
omit_actions = {
    'api',
    'sh',
    'os',
    'sql',
    'stack',
    'reload',
    'repo',
    'instance',
}
_paths = {
    'login'   : pages.login.layout,
    ''        : pages.dashboard.layout,
    'plugins' : pages.plugins.layout,
    'register': pages.register.layout,
}
_required_login = {''}
 
@dash_app.callback(
    Output('page-layout-div', 'children'),
    Output('session-store', 'data'),
    Input('location', 'pathname'),
    Input('session-store', 'data'),
    State('location', 'href'),
)
def update_page_layout_div(
        pathname: str, 
        session_store_data: Dict[str, Any],
        location_href: str,
    ) -> Tuple[List[Any], Dict[str, Any]]:
    """
    Route the user to the correct page.

    Parameters
    ----------
    pathname: str :
        The path in the browser.
        
    session_store_data: Dict[str, Any]:
        The stored session data.

    Returns
    -------
    A tuple of the page layout and new session store data.
    """
    ctx = dash.callback_context
    dash_endpoint = endpoints['dash']
    try:
        session_id = session_store_data.get('session-id', None) 
    except AttributeError:
        session_id = None

    ### Bypass login if `--no-auth` is specified.
    if session_id not in active_sessions and no_auth:
        session_store_data['session-id'] = str(uuid.uuid4())
        active_sessions[session_store_data['session-id']] = {'username': 'no-auth'}

        ### Sometimes the href is an empty string, so store it here for later.
        session_store_data['location.href'] = location_href
        session_store_to_return = session_store_data
    else:
        session_store_to_return = dash.no_update

    _path = (
        pathname.rstrip('/') + '/'
    ).replace(
        (dash_endpoint + '/'),
        ''
    ).rstrip('/').split('/')[0]
    path = (
        _path
        if no_auth or _path not in _required_login else (
            _path
            if session_id in active_sessions
            else 'login'
        )
    )
    layout = _paths.get(path, pages.error.layout)
    return layout, session_store_to_return


@dash_app.callback(
    Output('content-div-right', 'children'),
    Output('success-alert-div', 'children'),
    Output('webterm-div', 'children'),
    Output('webterm-div', 'style'),
    Input('go-button', 'n_clicks'),
    Input('get-pipes-button', 'n_clicks'),
    Input('get-jobs-button', 'n_clicks'),
    Input('get-plugins-button', 'n_clicks'),
    Input('get-users-button', 'n_clicks'),
    Input('get-graphs-button', 'n_clicks'),
    State('location', 'href'),
    State('session-store', 'data'),
    State('webterm-div', 'children'),
    *keys_state,
)
def update_content(*args):
    """
    Determine which trigger is seeking to update the content div,
    and execute the appropriate function.
    """
    ctx = dash.callback_context
    location_href = ctx.states['session-store.data'].get('location.href', None)
    session_id = ctx.states['session-store.data'].get('session-id', None)
    authenticated = is_session_authenticated(str(session_id))

    trigger = None
    initial_load = False
    ### Open the webterm on the initial load.
    if not ctx.triggered:
        initial_load = True
        trigger = 'open-shell-button'

    trigger = ctx.triggered[0]['prop_id'].split('.')[0] if not trigger else trigger

    session_data = args[-1]

    ### NOTE: functions MUST return a list of content and a list of alerts
    triggers = {
        'go-button': lambda x: ([], []),
        'cancel-button': lambda x: ([], []),
        'get-pipes-button': get_pipes_cards,
        'get-jobs-button': get_jobs_cards,
        'get-plugins-button': get_plugins_cards,
        'get-users-button': get_users_cards,
        'get-graphs-button': get_graphs_cards,
        'open-shell-button': lambda x: ([], []),
    }
    ### Defaults to 3 if not in dict.
    trigger_num_cols = {
        'get-graphs-button': 1,
        'get-pipes-button': 1,
        'get-jobs-button': 2,
    }
    
    ### NOTE: stop the running action if it exists
    stop_action(ctx.states)

    content, alerts = triggers[trigger](
        ctx.states,
        **filter_keywords(triggers[trigger], session_data=session_data)
    )
    webterm_style = {
        'display': (
            'none'
            if trigger not in ('open-shell-button', 'cancel-button', 'go-button')
            else 'block'
        )
    }

    ### If the webterm fails on initial load (e.g. insufficient permissions),
    ### don't display the alerts just yet.
    webterm_loaded = ctx.states.get('webterm-div.children', None) is not None
    if initial_load or not webterm_loaded or not authenticated:
        webterm, webterm_alerts = get_webterm(ctx.states)
        if webterm_style['display'] == 'block':
            alerts.extend(webterm_alerts)
    else:
        webterm = dash.no_update

    if initial_load and alerts:
        return console_div, [], [], {'display': 'none'}

    if trigger.startswith('get-') and trigger.endswith('-button'):
        content = build_cards_grid(content, num_columns=trigger_num_cols.get(trigger, 3))
    return content, alerts, webterm, webterm_style


dash_app.clientside_callback(
    """
    function(
        n_clicks,
        url,
        connector_keys,
        metric_keys,
        location_keys,
        flags,
        input_flags,
        input_flags_texts,
        instance,
    ){
        if (!n_clicks){ return url; }
        iframe = document.getElementById('webterm-iframe');
        if (!iframe){ return url; }

        // Actions must be obtained from the DOM because of dynamic subactions.
        action = document.getElementById('action-dropdown').value;
        subaction = document.getElementById('subaction-dropdown').value;
        subaction_text = document.getElementById('subaction-dropdown-text').value;

        iframe.contentWindow.postMessage(
            {
                action: action,
                subaction: subaction,
                subaction_text: subaction_text,
                connector_keys: connector_keys,
                metric_keys: metric_keys,
                location_keys: location_keys,
                flags: flags,
                input_flags: input_flags,
                input_flags_texts: input_flags_texts,
                instance: instance,
            },
            url
        );
        return url;
    }
    """,
    Output('location', 'href'),
    Input('go-button', 'n_clicks'),
    State('location', 'href'),
    State('connector-keys-dropdown', 'value'),
    State('metric-keys-dropdown', 'value'),
    State('location-keys-dropdown', 'value'),
    State('flags-dropdown', 'value'),
    State({'type': 'input-flags-dropdown', 'index': ALL}, 'value'),
    State({'type': 'input-flags-dropdown-text', 'index': ALL}, 'value'),
    State('instance-select', 'value'),
)

@dash_app.callback(
    Output('action-dropdown', 'value'),
    Output('subaction-dropdown', 'value'),
    Output('action-dropdown', 'options'),
    Output('subaction-dropdown', 'options'),
    Output('subaction-dropdown-div', 'hidden'),
    Input('action-dropdown', 'value'),
    Input('subaction-dropdown', 'value'),
)
def update_actions(action: str, subaction: str):
    """
    Update the subactions dropdown to reflect options for the primary action.
    """
    if not action:
        action, subaction = 'show', 'pipes'
        trigger = None
    _actions_options = sorted([
        {
            'label': a.replace('_', ' '),
            'value': a,
            'title': (textwrap.dedent(f.__doc__).lstrip() if f.__doc__ else 'No help available.'),
        }
        for a, f in actions.items() if a not in omit_actions
    ], key=lambda k: k['label'])
    _subactions_options = sorted([
        {
            'label': sa.replace('_', ' '),
            'value': sa,
            'title': (textwrap.dedent(f.__doc__).lstrip() if f.__doc__ else 'No help available.'),
        }
        for sa, f in get_subactions(action).items()
    ], key=lambda k: k['label'])
    _subactions = [o['label'] for o in _subactions_options]
    if subaction is None:
        subaction = _subactions[0] if _subactions else ''

    return (
        action,
        subaction,
        _actions_options,
        _subactions_options,
        len(_subactions) == 0,
    )


@dash_app.callback(
    Output('input-flags-div', 'children'),
    Output('flags-dropdown', 'options'),
    Input({'type': 'input-flags-dropdown', 'index': ALL}, 'value'),
    Input({'type': 'input-flags-remove-button', 'index': ALL}, 'n_clicks'),
    State({'type': 'input-flags-dropdown-text', 'index': ALL}, 'value'),
)
def update_flags(input_flags_dropdown_values, n_clicks, input_flags_texts):
    """
    Update the flags dropdowns on updates.
    """
    ctx = dash.callback_context
    trigger = ctx.triggered[0]['prop_id'].split('.')[0]
    trigger_dict = json.loads(trigger) if trigger else {}
    trigger_type = trigger_dict.get('type', None)
    taken_input_flags = set(flatten_list(input_flags_dropdown_values))

    def build_flags_options(is_input: bool = False):
        _flags_options = []
        for a in parser._actions:
            acceptable_args = (a.nargs != 0 if not is_input else a.nargs == 0)
            if acceptable_args or a.dest in omit_flags:
                continue
            _op = {'title': a.help}
            for _trigger in a.option_strings:
                if _trigger.startswith('--'):
                    _op['value'] = _trigger
                    break
            if not _op.get('value', None):
                _op['value'] = a.dest
            _op['label'] = _op['value']
            _flags_options.append(_op)
        return sorted(_flags_options, key=lambda k: k['label'])

    def build_row(index: int, val: Optional[str], val_text: Optional[str]):
        options = [
            op
            for op in build_flags_options(is_input=True)
            if op['value'] == val or op['value'] not in taken_input_flags
        ]
        row_children = [
            dbc.Col(
                html.Div(
                    dbc.InputGroup([
                        dbc.Button(
                            '❌',
                            color = 'link',
                            id = {'type': 'input-flags-remove-button', 'index': index},
                            size = 'sm',
                            style = {'text-decoration': 'none'},
                        ),
                        dcc.Dropdown(
                            id = {'type': 'input-flags-dropdown', 'index': index},
                            multi = False,
                            placeholder = 'Input flags',
                            options = options,
                            value = val,
                            style = {'flex': 1},
                        ),
                    ]),
                    id = {'type': 'input-flags-dropdown-div', 'index': index},
                    className = 'dbc_dark',
                ),
                sm = 12,
                md = 5,
                lg = 5,
                id = 'input-flags-left-col',
            ),
            dbc.Col(
                html.Div(
                    dbc.Input(
                        id = {'type': 'input-flags-dropdown-text', 'index': index},
                        placeholder = 'Flag value',
                        className = 'input-text',
                        value = val_text,
                    ),
                    id = {'type': 'input-flags-text-div', 'index': index},
                    className = 'dbc_dark input-text',
                ),
                sm = 12,
                md = 7,
                lg = 7,
            )
        ]
        return dbc.Row(
            row_children,
            id = {'type': 'input-flags-row', 'index': index},
            className = 'input-text',
        )

    rows = [
        build_row(i, val, val_text)
        for i, (val, val_text) in enumerate(zip(input_flags_dropdown_values, input_flags_texts))
    ]

    if trigger_type == 'input-flags-remove-button':
        remove_index = trigger_dict['index']
        try:
            del rows[remove_index]
        except IndexError:
            pass

    if not rows or input_flags_dropdown_values[-1]:
        rows.append(build_row(len(rows), None, None))

    return rows, build_flags_options()


@dash_app.callback(
    Output(component_id='connector-keys-dropdown', component_property='options'),
    Output(component_id='connector-keys-list', component_property='children'),
    Output(component_id='connector-keys-dropdown', component_property='value'),
    Output(component_id='metric-keys-dropdown', component_property='options'),
    Output(component_id='metric-keys-list', component_property='children'),
    Output(component_id='metric-keys-dropdown', component_property='value'),
    Output(component_id='location-keys-dropdown', component_property='options'),
    Output(component_id='location-keys-list', component_property='children'),
    Output(component_id='location-keys-dropdown', component_property='value'),
    Output(component_id='instance-select', component_property='value'),
    Output(component_id='instance-alert-div', component_property='children'),
    Input(component_id='connector-keys-dropdown', component_property='value'),
    Input(component_id='metric-keys-dropdown', component_property='value'),
    Input(component_id='location-keys-dropdown', component_property='value'),
    Input(component_id='instance-select', component_property='value'),
    *keys_state
)
def update_keys_options(
        connector_keys: Optional[List[str]],
        metric_keys: Optional[List[str]],
        location_keys: Optional[List[str]],
        instance_keys: Optional[str],
        *keys
    ):
    """
    Update the keys dropdown menus' options.
    """
    ctx = dash.callback_context
    trigger = ctx.triggered[0]['prop_id'].split('.')[0]
    instance_click = trigger == 'instance-select'

    ### Update the instance first.
    update_instance_keys = False
    if not instance_keys:
        instance_keys = str(get_api_connector())
        update_instance_keys = True
    instance_alerts = []
    try:
        parse_instance_keys(instance_keys)
    except Exception as e:
        instance_alerts += [alert_from_success_tuple((False, str(e)))]

    ### Update the keys filters.
    if connector_keys is None:
        connector_keys = []
    if metric_keys is None:
        metric_keys = []
    if location_keys is None:
        location_keys = []
    num_filter = 0
    if connector_keys:
        num_filter += 1
    if metric_keys:
        num_filter += 1
    if location_keys:
        num_filter += 1

    _ck_alone, _mk_alone, _lk_alone = False, False, False
    _ck_filter, _mk_filter, _lk_filter = connector_keys, metric_keys, location_keys

    _ck_alone = (connector_keys and num_filter == 1) or instance_click
    _mk_alone = (metric_keys and num_filter == 1) or instance_click
    _lk_alone = (location_keys and num_filter == 1) or instance_click

    from meerschaum.utils import fetch_pipes_keys

    try:
        _all_keys = fetch_pipes_keys('registered', get_web_connector(ctx.states))
        _keys = fetch_pipes_keys(
            'registered', get_web_connector(ctx.states),
            connector_keys=_ck_filter, metric_keys=_mk_filter, location_keys=_lk_filter
        )
    except Exception as e:
        instance_alerts += [alert_from_success_tuple((False, str(e)))]
        _all_keys, _keys = [], []
    _connectors_options = []
    _metrics_options = []
    _locations_options = []

    _seen_keys = {'ck' : set(), 'mk' : set(), 'lk' : set()}

    def add_options(options, keys, key_type):
        for ck, mk, lk in keys:
            k = locals()[key_type]
            if k not in _seen_keys[key_type]:
                _k = 'None' if k in (None, '[None]', 'None', 'null') else k
                options.append({'label': _k, 'value': _k})
                _seen_keys[key_type].add(k)

    add_options(_connectors_options, _all_keys if _ck_alone else _keys, 'ck')
    add_options(_metrics_options, _all_keys if _mk_alone else _keys, 'mk')
    add_options(_locations_options, _all_keys if _lk_alone else _keys, 'lk')
    connector_keys = sorted([
        ck
        for ck in connector_keys 
        if ck in [
            _ck['value']
            for _ck in _connectors_options
        ]
    ])
    metric_keys = sorted([
        mk
        for mk in metric_keys
        if mk in [
            _mk['value']
            for _mk in _metrics_options
        ]
    ])
    location_keys = sorted([
        lk
        for lk in location_keys
        if lk in [
            _lk['value']
            for _lk in _locations_options
        ]
    ])
    _connectors_datalist = [html.Option(value=o['value']) for o in _connectors_options]
    _metrics_datalist = [html.Option(value=o['value']) for o in _metrics_options]
    _locations_datalist = [html.Option(value=o['value']) for o in _locations_options]
    return (
        _connectors_options,
        _connectors_datalist,
        connector_keys,
        _metrics_options,
        _metrics_datalist,
        metric_keys,
        _locations_options,
        _locations_datalist,
        location_keys,
        (instance_keys if update_instance_keys else dash.no_update),
        instance_alerts,
    )

dash_app.clientside_callback(
    """
    function(
        instance,
        url,
    ){
        window.instance = instance;
        if (!instance){ return url; }
        iframe = document.getElementById('webterm-iframe');
        if (!iframe){ return url; }
        if (!window.instance){
            window.instance = instance;
            return url;
        }
        window.instance = instance;

        iframe.contentWindow.postMessage(
            {
                action: "instance",
                subaction_text: instance,
            },
            url
        );
        return url;
    }
    """,
    Output('location', 'href'),
    Input('instance-select', 'value'),
)


@dash_app.callback(
    Output(component_id='connector-keys-input', component_property='value'),
    Input(component_id='clear-connector-keys-input-button', component_property='n_clicks'),
)
def clear_connector_keys_input(val):
    """
    Reset the connector key input box.
    """
    return ''

@dash_app.callback(
    Output(component_id='metric-keys-input', component_property='value'),
    Input(component_id='clear-metric-keys-input-button', component_property='n_clicks'),
)
def clear_metric_keys_input(val):
    """
    Reset the metric key input box.
    """
    return ''

@dash_app.callback(
    Output(component_id='location-keys-input', component_property='value'),
    Input(component_id='clear-location-keys-input-button', component_property='n_clicks'),
)
def clear_location_keys_input(val):
    """
    Reset the connector key input box.
    """
    return ''

@dash_app.callback(
    Output(component_id='subaction-dropdown-text', component_property='value'),
    Input(component_id='clear-subaction-dropdown-text-button', component_property='n_clicks'),
)
def clear_subaction_dropdown_text(val):
    """
    Reset the connector key input box.
    """
    return ''

@dash_app.callback(
    Output('arguments-collapse', 'is_open'),
    Input('show-arguments-collapse-button', 'n_clicks'),
    State('arguments-collapse', 'is_open'),
)
def show_arguments_collapse(n_clicks : int, is_open : bool):
    """
    Show or hide the arguments Collapse.
    """
    return not is_open if n_clicks else is_open


dash_app.clientside_callback(
    """
    function(console_children, url){
        if (!console_children){
            return console_children;
        }
        var ansi_up = new AnsiUp;
        var html = ansi_up.ansi_to_html(console_children);
        console_div = document.getElementById("console-div");
        console_div.innerHTML = (
            "<pre id=\\"console-pre\\">" + html + "</pre>"
        );
        console_div.scrollTop = console_div.scrollHeight;
        return url;
    }
    """,
    Output('location', 'href'),
    Input('console-pre', 'children'),
    State('location', 'href'),
)

@dash_app.callback(
    Output("download-dataframe-csv", "data"),
    Input({'type': 'pipe-download-csv-button', 'index': ALL}, 'n_clicks'),
)
def download_pipe_csv(n_clicks):
    """
    Download the most recent chunk as a CSV file.
    """
    if not n_clicks:
        raise PreventUpdate
    ctx = dash.callback_context.triggered
    if ctx[0]['value'] is None:
        raise PreventUpdate
    pipe = pipe_from_ctx(ctx, 'n_clicks')
    if pipe is None:
        raise PreventUpdate
    filename = str(pipe.target) + '.csv'
    bounds = pipe.get_chunk_bounds(bounded=True, debug=debug)
    begin, _ = bounds[-1]
    try:
        df = pipe.get_data(begin=begin, end=None, debug=debug)
    except Exception as e:
        df = None
    if df is not None:
        return dcc.send_data_frame(df.to_csv, filename, index=False)
    raise PreventUpdate


@dash_app.callback(
    Output({'type': 'pipe-accordion', 'index': MATCH}, 'children'),
    Input({'type': 'pipe-accordion', 'index': MATCH}, 'active_item'),
    State('session-store', 'data'),
)
def update_pipe_accordion(item, session_store_data):
    if item is None:
        raise PreventUpdate

    ctx = dash.callback_context.triggered
    if ctx[0]['value'] is None:
        raise PreventUpdate
    pipe = pipe_from_ctx(ctx, 'active_item')
    if pipe is None:
        raise PreventUpdate

    session_id = session_store_data.get('session-id', None)
    authenticated = is_session_authenticated(str(session_id))
    return accordion_items_from_pipe(pipe, active_items=[item], authenticated=authenticated)


@dash_app.callback(
    Output({'type': 'update-parameters-success-div', 'index': MATCH}, 'children'),
    Input({'type': 'update-parameters-button', 'index': MATCH}, 'n_clicks'),
    State({'type': 'parameters-editor', 'index': MATCH}, 'value')
)
def update_pipe_parameters_click(n_clicks, parameters_editor_text):
    if not n_clicks:
        raise PreventUpdate
    ctx = dash.callback_context
    triggered = dash.callback_context.triggered
    if triggered[0]['value'] is None:
        raise PreventUpdate
    pipe = pipe_from_ctx(triggered, 'n_clicks')
    if pipe is None:
        raise PreventUpdate

    if parameters_editor_text is None:
        success, msg = False, f"Unable to update parameters for {pipe}."
    else:
        try:
            text_format = 'JSON' if parameters_editor_text.lstrip().startswith('{') else 'YAML'
            params = (
                json.loads(parameters_editor_text)
                if text_format == 'JSON'
                else yaml.load(parameters_editor_text)
            )
            pipe.parameters = params
            success, msg = pipe.edit(debug=debug)
        except Exception as e:
            success, msg = False, f"Invalid {text_format}:\n{e}"

    return alert_from_success_tuple((success, msg))


@dash_app.callback(
    Output({'type': 'update-sql-success-div', 'index': MATCH}, 'children'),
    Input({'type': 'update-sql-button', 'index': MATCH}, 'n_clicks'),
    State({'type': 'sql-editor', 'index': MATCH}, 'value')
)
def update_pipe_sql_click(n_clicks, sql_editor_text):
    if not n_clicks:
        raise PreventUpdate
    ctx = dash.callback_context
    triggered = dash.callback_context.triggered
    if triggered[0]['value'] is None:
        raise PreventUpdate
    pipe = pipe_from_ctx(triggered, 'n_clicks')
    if pipe is None:
        raise PreventUpdate

    if sql_editor_text is None:
        success, msg = False, f"Unable to update SQL definition for {pipe}."
    else:
        try:
            set_pipe_query(pipe, sql_editor_text)
            success, msg = pipe.edit(debug=debug)
        except Exception as e:
            success, msg = False, f"Invalid SQL query:\n{e}"

    return alert_from_success_tuple((success, msg))


@dash_app.callback(
    Output({'type': 'sync-success-div', 'index': MATCH}, 'children'),
    Input({'type': 'update-sync-button', 'index': MATCH}, 'n_clicks'),
    State({'type': 'sync-editor', 'index': MATCH}, 'value')
)
def sync_documents_click(n_clicks, sync_editor_text):
    if not n_clicks:
        raise PreventUpdate
    ctx = dash.callback_context
    triggered = dash.callback_context.triggered
    if triggered[0]['value'] is None:
        raise PreventUpdate
    pipe = pipe_from_ctx(triggered, 'n_clicks')
    if pipe is None:
        raise PreventUpdate

    try:
        msg = '... '
        docs = json.loads(sync_editor_text)
    except Exception as e:
        docs = None
        msg = str(e)
    if docs is None:
        success, msg = False, (msg + f"Unable to sync documents to {pipe}.")
    else:
        try:
            success, msg = pipe.sync(docs, debug=debug)
        except Exception as e:
            import traceback
            traceback.print_exc()
            success, msg = False, f"Encountered exception:\n{e}"

    return alert_from_success_tuple((success, msg))


dash_app.clientside_callback(
    """
    function(n_clicks_arr, url){
        display_block = {"display": "block"};

        var clicked = false;
        for (var i = 0; i < n_clicks_arr.length; i++){
            if (n_clicks_arr[i]){
                clicked = true;
                break;
            }
        }
        if (!clicked){ return dash_clientside.no_update; }

        const triggered_id = dash_clientside.callback_context.triggered_id;
        const action = triggered_id["action"];
        const pipe_meta = JSON.parse(triggered_id["index"]);

        iframe = document.getElementById('webterm-iframe');
        if (!iframe){ return dash_clientside.no_update; }
        var location = pipe_meta.location;
        if (!pipe_meta.location){
            location = "None";
        }

        iframe.contentWindow.postMessage(
            {
                action: action,
                subaction: "pipes",
                connector_keys: [pipe_meta.connector],
                metric_keys: [pipe_meta.metric],
                location_keys: [location],
                instance: pipe_meta.instance,
            },
            url
        );
        dash_clientside.set_props("webterm-div", {"style": display_block});
        return [];
    }
    """,
    Output('content-div-right', 'children'),
    Input({'type': 'manage-pipe-button', 'index': ALL, 'action': ALL}, 'n_clicks'),
    State('location', 'href'),
)

@dash_app.callback(
    Output("navbar-collapse", "is_open"),
    [Input("navbar-toggler", "n_clicks")],
    [State("navbar-collapse", "is_open")],
)
def toggle_navbar_collapse(n_clicks: Optional[int], is_open: bool) -> bool:
    """
    Toggle the Instance selection collapse in the navbar.

    Parameters
    ----------
    n_clicks: Optional[int]
        The number of times the toggler was clicked.

    is_open: bool
        The current state of the collapse.

    Returns
    -------
    The toggled state of the collapse.
    """
    if n_clicks:
        return not is_open
    return is_open


@dash_app.callback(
    Output('location', 'pathname'),
    Output('session-store', 'data'),
    Input("sign-out-button", "n_clicks"),
    State('session-store', 'data'),
)
def sign_out_button_click(
        n_clicks: Optional[int],
        session_store_data: Dict[str, Any],
    ):
    """
    When the sign out button is clicked, remove the session data and redirect to the login page.
    """
    if not n_clicks:
        raise PreventUpdate
    session_id = session_store_data.get('session-id', None)
    if session_id:
        _ = active_sessions.pop(session_id, None)
        _ = authenticated_sessions.pop(session_id, None)
        _ = unauthenticated_sessions.pop(session_id, None)
    return endpoints['dash'], {}


@dash_app.callback(
    Output({'type': 'parameters-editor', 'index': MATCH}, 'value'),
    Input({'type': 'parameters-as-yaml-button', 'index': MATCH}, 'n_clicks'),
    Input({'type': 'parameters-as-json-button', 'index': MATCH}, 'n_clicks'),
)
def parameters_as_yaml_or_json_click(
        yaml_n_clicks: Optional[int],
        json_n_clicks: Optional[int],
    ):
    """
    When the `YAML` button is clicked under the parameters editor, switch the content to YAML.
    """
    if not yaml_n_clicks and not json_n_clicks:
        raise PreventUpdate

    ctx = dash.callback_context
    triggered = dash.callback_context.triggered
    if triggered[0]['value'] is None:
        raise PreventUpdate
    as_yaml = 'yaml' in triggered[0]['prop_id']
    pipe = pipe_from_ctx(triggered, 'n_clicks')
    if pipe is None:
        raise PreventUpdate

    if as_yaml:
        return yaml.dump(pipe.parameters)
    return json.dumps(pipe.parameters, indent=4, separators=(',', ': '), sort_keys=True)



#! /usr/bin/env python3
# -*- coding: utf-8 -*-
# vim:fenc=utf-8

"""
Routes to the Webterm proxy.
"""

import asyncio
from meerschaum.utils.typing import Optional
from meerschaum.api import app, no_auth, manager, endpoints
from meerschaum.utils.packages import attempt_import
from meerschaum.api.dash import active_sessions
from meerschaum.api.dash.users import is_session_authenticated
fastapi, fastapi_responses = attempt_import('fastapi', 'fastapi.responses')
import starlette
httpx = attempt_import('httpx')
websockets = attempt_import('websockets')
Request = fastapi.Request
WebSocket = fastapi.WebSocket
HTMLResponse = fastapi_responses.HTMLResponse
Response = fastapi_responses.Response
PlainTextResponse = fastapi_responses.PlainTextResponse


@app.get(endpoints['webterm'], tags=["Webterm"])
async def get_webterm(
        request: Request,
        s: Optional[str] = None,
    ) -> HTMLResponse:
    """
    Get the main HTML template for the Webterm.
    """
    if not is_session_authenticated(s):
        return HTMLResponse(
            """
            <html>
                <head>
                    <style>
                    body {
                        background-color: #222222;
                    }
                    #message-div {
                        padding: 15px;
                        font-size: x-large;
                        font-family: sans-serif;
                    }
                    h2 {
                        color: #FFFFFF;
                    }
                    p {
                        color: #FFFFFF;
                    }
                    a {
                        color: #FFFFFF;
                        text-decoration: underline;
                    }
                    </style>
                </head>
                <body>
                    <div id="message-div">
                        <h2>Unauthorized</h2>
                        <p>Please <a href="/dash/login">log in</a> and try again.</p>
                    </div>
                </body>
            </html>
            """,
            status_code = 401,
        )

    async with httpx.AsyncClient() as client:
        response = await client.get("http://localhost:8765/")
        text = response.text
        if request.url.scheme == 'https':
            text = text.replace('ws://', 'wss://')
        return HTMLResponse(
            content = text,
            status_code = response.status_code,
            headers = request.headers,
        )


@app.websocket(endpoints['webterm_websocket'])
async def webterm_websocket(websocket: WebSocket):
    """
    Connect to the Webterm's websocket.
    """
    try:
        await websocket.accept()
        session_doc = await websocket.receive_json()
    except starlette.websockets.WebSocketDisconnect:
        return
    session_id = (session_doc or {}).get('session-id', 'no-auth')

    if not is_session_authenticated(session_id):
        await websocket.close()
        return

    async with websockets.connect("ws://localhost:8765/websocket") as ws:
        async def forward_messages():
            try:
                while True:
                    client_message = await websocket.receive_text()
                    await ws.send(client_message)
            except (
                websockets.exceptions.ConnectionClosed,
                starlette.websockets.WebSocketDisconnect
            ):
                await ws.close()

        async def backward_messages():
            try:
                while True:
                    server_response = await ws.recv()
                    await websocket.send_text(server_response)
            except (
                websockets.exceptions.ConnectionClosed,
                starlette.websockets.WebSocketDisconnect
            ):
                pass

        await asyncio.gather(
            forward_messages(),
            backward_messages(),
        )



#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:fenc=utf-8

"""
Interface with SQL servers using sqlalchemy.
"""

from __future__ import annotations
import meerschaum as mrsm
from meerschaum.utils.typing import Optional, Any, Union

from meerschaum.connectors import Connector
from meerschaum.utils.warnings import error

class SQLConnector(Connector):
    """
    Connect to SQL databases via `sqlalchemy`.
    
    SQLConnectors may be used as Meerschaum instance connectors.
    Read more about connectors and instances at
    https://meerschaum.io/reference/connectors/

    """

    IS_INSTANCE: bool = True

    from ._create_engine import flavor_configs, create_engine
    from ._sql import read, value, exec, execute, to_sql, exec_queries
    from meerschaum.utils.sql import test_connection
    from ._fetch import fetch, get_pipe_metadef
    from ._cli import cli, _cli_exit
    from ._pipes import (
        fetch_pipes_keys,
        create_indices,
        drop_indices,
        get_create_index_queries,
        get_drop_index_queries,
        get_add_columns_queries,
        get_alter_columns_queries,
        delete_pipe,
        get_pipe_data,
        get_pipe_data_query,
        register_pipe,
        edit_pipe,
        get_pipe_id,
        get_pipe_attributes,
        sync_pipe,
        sync_pipe_inplace,
        get_sync_time,
        pipe_exists,
        get_pipe_rowcount,
        drop_pipe,
        clear_pipe,
        deduplicate_pipe,
        get_pipe_table,
        get_pipe_columns_types,
        get_to_sql_dtype,
        get_pipe_schema,
    )
    from ._plugins import (
        register_plugin,
        delete_plugin,
        get_plugin_id,
        get_plugin_version,
        get_plugins,
        get_plugin_user_id,
        get_plugin_username,
        get_plugin_attributes,
    )
    from ._users import (
        register_user,
        get_user_id,
        get_users,
        edit_user,
        delete_user,
        get_user_password_hash,
        get_user_type,
        get_user_attributes,
    )
    from ._uri import from_uri, parse_uri
    from ._instance import (
        _log_temporary_tables_creation,
        _drop_temporary_table,
        _drop_temporary_tables,
        _drop_old_temporary_tables,
    )
    
    def __init__(
        self,
        label: Optional[str] = None,
        flavor: Optional[str] = None,
        wait: bool = False,
        connect: bool = False,
        debug: bool = False,
        **kw: Any
    ):
        """
        Parameters
        ----------
        label: str, default 'main'
            The identifying label for the connector.
            E.g. for `sql:main`, 'main' is the label.
            Defaults to 'main'.

        flavor: Optional[str], default None
            The database flavor, e.g.
            `'sqlite'`, `'postgresql'`, `'cockroachdb'`, etc.
            To see supported flavors, run the `bootstrap connectors` command.

        wait: bool, default False
            If `True`, block until a database connection has been made.
            Defaults to `False`.

        connect: bool, default False
            If `True`, immediately attempt to connect the database and raise
            a warning if the connection fails.
            Defaults to `False`.

        debug: bool, default False
            Verbosity toggle.
            Defaults to `False`.

        kw: Any
            All other arguments will be passed to the connector's attributes.
            Therefore, a connector may be made without being registered,
            as long enough parameters are supplied to the constructor.
        """
        if 'uri' in kw:
            uri = kw['uri']
            if uri.startswith('postgres://'):
                uri = uri.replace('postgres://', 'postgresql://', 1)
            if uri.startswith('timescaledb://'):
                uri = uri.replace('timescaledb://', 'postgresql://', 1)
                flavor = 'timescaledb'
            kw['uri'] = uri
            from_uri_params = self.from_uri(kw['uri'], as_dict=True)
            label = label or from_uri_params.get('label', None)
            _ = from_uri_params.pop('label', None)

            ### Sometimes the flavor may be provided with a URI.
            kw.update(from_uri_params)
            if flavor:
                kw['flavor'] = flavor


        ### set __dict__ in base class
        super().__init__(
            'sql',
            label = label or self.__dict__.get('label', None),
            **kw
        )

        if self.__dict__.get('flavor', None) == 'sqlite':
            self._reset_attributes()
            self._set_attributes(
                'sql',
                label = label,
                inherit_default = False,
                **kw
            )
            ### For backwards compatability reasons, set the path for sql:local if its missing.
            if self.label == 'local' and not self.__dict__.get('database', None):
                from meerschaum.config._paths import SQLITE_DB_PATH
                self.database = str(SQLITE_DB_PATH)

        ### ensure flavor and label are set accordingly
        if 'flavor' not in self.__dict__:
            if flavor is None and 'uri' not in self.__dict__:
                raise Exception(
                    f"    Missing flavor. Provide flavor as a key for '{self}'."
                )
            self.flavor = flavor or self.parse_uri(self.__dict__['uri']).get('flavor', None)

        if self.flavor == 'postgres':
            self.flavor = 'postgresql'

        self._debug = debug
        ### Store the PID and thread at initialization
        ### so we can dispose of the Pool in child processes or threads.
        import os, threading
        self._pid = os.getpid()
        self._thread_ident = threading.current_thread().ident
        self._sessions = {}
        self._locks = {'_sessions': threading.RLock(), }

        ### verify the flavor's requirements are met
        if self.flavor not in self.flavor_configs:
            error(f"Flavor '{self.flavor}' is not supported by Meerschaum SQLConnector")
        if not self.__dict__.get('uri'):
            self.verify_attributes(
                self.flavor_configs[self.flavor].get('requirements', set()),
                debug=debug,
            )

        if wait:
            from meerschaum.connectors.poll import retry_connect
            retry_connect(connector=self, debug=debug)

        if connect:
            if not self.test_connection(debug=debug):
                from meerschaum.utils.warnings import warn
                warn(f"Failed to connect with connector '{self}'!", stack=False)

    @property
    def Session(self):
        if '_Session' not in self.__dict__:
            if self.engine is None:
                return None

            from meerschaum.utils.packages import attempt_import
            sqlalchemy_orm = attempt_import('sqlalchemy.orm')
            session_factory = sqlalchemy_orm.sessionmaker(self.engine)
            self._Session = sqlalchemy_orm.scoped_session(session_factory)

        return self._Session

    @property
    def engine(self):
        import os, threading
        ### build the sqlalchemy engine
        if '_engine' not in self.__dict__:
            self._engine, self._engine_str = self.create_engine(include_uri=True)

        same_process = os.getpid() == self._pid
        same_thread = threading.current_thread().ident == self._thread_ident

        ### handle child processes
        if not same_process:
            self._pid = os.getpid()
            self._thread = threading.current_thread()
            from meerschaum.utils.warnings import warn
            warn(f"Different PID detected. Disposing of connections...")
            self._engine.dispose()

        ### handle different threads
        if not same_thread:
            pass

        return self._engine

    @property
    def DATABASE_URL(self) -> str:
        """
        Return the URI connection string (alias for `SQLConnector.URI`.
        """
        _ = self.engine
        return str(self._engine_str)

    @property
    def URI(self) -> str:
        """
        Return the URI connection string.
        """
        _ = self.engine
        return str(self._engine_str)

    @property
    def IS_THREAD_SAFE(self) -> str:
        """
        Return whether this connector may be multithreaded.
        """
        if self.flavor == 'duckdb':
            return False
        if self.flavor == 'sqlite':
            return ':memory:' not in self.URI
        return True


    @property
    def metadata(self):
        """
        Return the metadata bound to this configured schema.
        """
        from meerschaum.utils.packages import attempt_import
        sqlalchemy = attempt_import('sqlalchemy')
        if '_metadata' not in self.__dict__:
            self._metadata = sqlalchemy.MetaData(schema=self.schema)
        return self._metadata


    @property
    def instance_schema(self):
        """
        Return the schema name for Meerschaum tables. 
        """
        return self.schema


    @property
    def internal_schema(self):
        """
        Return the schema name for internal tables. 
        """
        from meerschaum.config.static import STATIC_CONFIG
        from meerschaum.utils.packages import attempt_import
        from meerschaum.utils.sql import NO_SCHEMA_FLAVORS
        schema_name = self.__dict__.get('internal_schema', None) or (
            STATIC_CONFIG['sql']['internal_schema']
            if self.flavor not in NO_SCHEMA_FLAVORS
            else self.schema
        )

        if '_internal_schema' not in self.__dict__:
            self._internal_schema = schema_name
        return self._internal_schema


    @property
    def db(self) -> Optional[databases.Database]:
        from meerschaum.utils.packages import attempt_import
        databases = attempt_import('databases', lazy=False, install=True)
        url = self.DATABASE_URL
        if 'mysql' in url:
            url = url.replace('+pymysql', '')
        if '_db' not in self.__dict__:
            try:
                self._db = databases.Database(url)
            except KeyError:
                ### Likely encountered an unsupported flavor.
                from meerschaum.utils.warnings import warn
                self._db = None
        return self._db


    @property
    def db_version(self) -> Union[str, None]:
        """
        Return the database version.
        """
        _db_version = self.__dict__.get('_db_version', None)
        if _db_version is not None:
            return _db_version

        from meerschaum.utils.sql import get_db_version
        self._db_version = get_db_version(self)
        return self._db_version


    @property
    def schema(self) -> Union[str, None]:
        """
        Return the default schema to use.
        A value of `None` will not prepend a schema.
        """
        if 'schema' in self.__dict__:
            return self.__dict__['schema']

        from meerschaum.utils.sql import NO_SCHEMA_FLAVORS
        if self.flavor in NO_SCHEMA_FLAVORS:
            self.__dict__['schema'] = None
            return None

        sqlalchemy = mrsm.attempt_import('sqlalchemy')
        _schema = sqlalchemy.inspect(self.engine).default_schema_name
        self.__dict__['schema'] = _schema
        return _schema


    def __getstate__(self):
        return self.__dict__

    def __setstate__(self, d):
        self.__dict__.update(d)

    def __call__(self):
        return self
